"""Views for the CircleCI integration (webhook listener)."""

import json
import logging

from django.http import HttpResponse, HttpResponseBadRequest
from django.views.generic import View
from reviewboard.reviews.models.status_update import StatusUpdate


logger = logging.getLogger(__name__)


class CircleCIWebHookView(View):
    """The view to handle webhook notifications from a CircleCI build."""

    STATUS_STATE_MAP = {
        'canceled': StatusUpdate.DONE_FAILURE,
        'infrastructure_fail': StatusUpdate.ERROR,
        'failed': StatusUpdate.DONE_FAILURE,
        'fixed': StatusUpdate.DONE_SUCCESS,
        'success': StatusUpdate.DONE_SUCCESS,
        'queued': StatusUpdate.PENDING,
        'running': StatusUpdate.PENDING,
        'scheduled': StatusUpdate.PENDING,
        'timedout': StatusUpdate.TIMEOUT,
    }

    STATUS_DESCRIPTION_MAP = {
        'canceled': 'build canceled.',
        'infrastructure_fail': 'build infrastructure failure.',
        'failed': 'build failed.',
        'fixed': 'build succeeded.',
        'success': 'build succeeded.',
        'queued': 'build queued.',
        'running': 'build running.',
        'scheduled': 'build scheduled.',
        'timedout': 'build running.',
    }

    def post(self, request, *args, **kwargs):
        """Handle the POST.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple):
                Additional positional arguments, parsed from the URL.

            **kwargs (dict):
                Additional keyword arguments, parsed from the URL.

        Returns:
            django.http.HttpResponse:
            A response.
        """
        payload = json.loads(request.body)

        try:
            build_parameters = payload['payload']['build_parameters']
        except KeyError:
            error = 'Unable to find build_parameters in payload.'
            logger.error('CircleCI webhook: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        try:
            review_request_id = \
                build_parameters['REVIEWBOARD_REVIEW_REQUEST']
            local_site_name = \
                build_parameters.get('REVIEWBOARD_LOCAL_SITE')
            status_update_id = \
                build_parameters['REVIEWBOARD_STATUS_UPDATE_ID']

            if local_site_name:
                local_site_log = ' (local_site %s)' % local_site_name
            else:
                local_site_log = ''
        except KeyError:
            # This was a normal build, not a review request.
            return HttpResponse()

        logger.debug('Got CircleCI webhook event for review request %d%s '
                     '(status update %d)',
                     review_request_id, local_site_log, status_update_id,
                     extra={'request': request})

        try:
            status_update = StatusUpdate.objects.get(pk=status_update_id)
        except StatusUpdate.DoesNotExist:
            error = ('Unable to find matching status update ID %d'
                     % status_update_id)
            logger.error('CircleCI webhook: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        status = payload['payload']['status']

        try:
            status_update.state = self.STATUS_STATE_MAP[status]
            status_update.description = self.STATUS_DESCRIPTION_MAP[status]
            status_update.save()
        except KeyError:
            # There are a few other possibilities for the "status" field but
            # we don't care about them.
            pass

        return HttpResponse()



"""Review Board extension for common integrations."""

from __future__ import annotations

from django.urls import include, path
from django.utils.translation import gettext_lazy as _
from reviewboard.extensions.base import Extension
from reviewboard.extensions.hooks import IntegrationHook, URLHook
from reviewboard.urls import reviewable_url_names, review_request_url_names


class RBIntegrationsExtension(Extension):
    """Extends Review Board with support for many common integrations."""

    metadata = {
        'Name': _('Review Board Integrations'),
        'Summary': _('A set of third-party service integrations for '
                     'Review Board.'),
    }

    css_bundles = {
        'fields': {
            'source_filenames': [
                'css/asana/asana.less',
                'css/trello/trello.less',
            ],
            'apply_to': reviewable_url_names + review_request_url_names,
        },
        'asana-integration-config': {
            'source_filenames': ['css/asana/integration-config.less'],
        },
        'travis-ci-integration-config': {
            'source_filenames': ['css/travisci/integration-config.less'],
        },
    }

    js_bundles = {
        'fields': {
            'source_filenames': [
                'js/fields/index.ts',
                'js/trello/trelloFieldView.es6.js',
            ],
            'apply_to': reviewable_url_names + review_request_url_names,
        },
        'asana-integration-config': {
            'source_filenames': [
                'js/asana/integrationConfig.ts',
            ],
        },
        'travis-ci-integration-config': {
            'source_filenames': ['js/travisci/integrationConfig.es6.js'],
        },
    }

    def initialize(self) -> None:
        """Initialize the extension."""
        from rbintegrations.asana.integration import AsanaIntegration
        from rbintegrations.circleci.integration import CircleCIIntegration
        from rbintegrations.discord.integration import DiscordIntegration
        from rbintegrations.gitlabci.integration import GitLabCIIntegration
        from rbintegrations.idonethis.integration import IDoneThisIntegration
        from rbintegrations.jenkinsci.integration import JenkinsCIIntegration
        from rbintegrations.matrix.integration import MatrixIntegration
        from rbintegrations.mattermost.integration import MattermostIntegration
        from rbintegrations.msteams.integration import MSTeamsIntegration
        from rbintegrations.slack.integration import SlackIntegration
        from rbintegrations.travisci.integration import TravisCIIntegration
        from rbintegrations.trello.integration import TrelloIntegration

        IntegrationHook(self, AsanaIntegration)
        IntegrationHook(self, CircleCIIntegration)
        IntegrationHook(self, DiscordIntegration)
        IntegrationHook(self, GitLabCIIntegration)
        IntegrationHook(self, IDoneThisIntegration)
        IntegrationHook(self, JenkinsCIIntegration)
        IntegrationHook(self, MatrixIntegration)
        IntegrationHook(self, MattermostIntegration)
        IntegrationHook(self, MSTeamsIntegration)
        IntegrationHook(self, SlackIntegration)
        IntegrationHook(self, TravisCIIntegration)
        IntegrationHook(self, TrelloIntegration)

        URLHook(self, [
            path('rbintegrations/asana/',
                 include('rbintegrations.asana.urls')),
            path('rbintegrations/circle-ci/',
                 include('rbintegrations.circleci.urls')),
            path('rbintegrations/gitlab-ci/',
                 include('rbintegrations.gitlabci.urls')),
            path('rbintegrations/travis-ci/',
                 include('rbintegrations.travisci.urls')),
            path('rbintegrations/trello/',
                 include('rbintegrations.trello.urls')),
        ])



"""Utilities for interacting with the Jenkins CI API."""

import json
import logging
from urllib.error import HTTPError
from urllib.parse import (quote, urlencode)
from urllib.request import urlopen

from reviewboard.hostingsvcs.service import HostingServiceHTTPRequest


logger = logging.getLogger(__name__)


class JenkinsAPI(object):
    """Object for interacting with the Jenkins CI API."""

    def __init__(self, endpoint, job_name, username, password):
        """Initialize the object.

        Args:
            endpoint (unicode):
                Jenkins server endpoint.

            job_name (unicode):
                Job name on Jenkins.

            username (unicode):
                Jenkins username.

            password (unicode):
                Jenkins password.
        """
        self.endpoint = endpoint
        self.job_name = job_name
        self.username = username
        self.password = password
        self.csrf_protection_enabled = True
        self.crumb = None
        self.crumb_request_field = None

    def test_connection(self):
        """Test the connection to the Jenkins server.

        This is used for verifying both the URL and user credentials are
        correct.
        """
        self._make_request('%s/api/json?pretty=true' % self.endpoint,
                           method='GET')

    def start_build(self, patch_info):
        """Start a build.

        Args:
            patch_info (dict):
                Contains the review ID, review branch, review diff revision
                and the status update ID.

        Raises:
            urllib2.URLError:
                The HTTP request failed.
        """
        data = {
            'parameter': [
                {
                    'name': 'REVIEWBOARD_SERVER',
                    'value': patch_info['reviewboard_server'],
                },
                {
                    'name': 'REVIEWBOARD_REPOSITORY_ID',
                    'value': patch_info['repository_id'],
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_BRANCH',
                    'value': patch_info['review_branch'],
                },
                {
                    'name': 'REVIEWBOARD_DIFF_REVISION',
                    'value': patch_info['diff_revision'],
                },
                {
                    'name': 'REVIEWBOARD_REVIEW_ID',
                    'value': patch_info['review_id'],
                },
                {
                    'name': 'REVIEWBOARD_STATUS_UPDATE_ID',
                    'value': patch_info['status_update_id'],
                },
            ],
        }

        # This is not part of the official REST API, but is however listed in
        # the Jenkins wiki as the correct way to initiate a remote build.
        #
        # This method of passing in the build parameters may change in the
        # future.
        self._make_request(
            '%s/job/%s/build' % (self.endpoint,
                                 quote(self.job_name)),
            body=urlencode({
                'json': json.dumps(data, sort_keys=True)
            }),
            content_type='application/x-www-form-urlencoded',
            method='POST'
        )

    def _fetch_csrf_token(self):
        """Fetches a CSRF token from the Jenkins server.

        This is required for making requests to API endpoints when using basic
        authentication. A crumb is no longer required when using API token
        authentication to access buildWithParameters.
        """
        data = self._make_raw_request('%s/crumbIssuer/api/json'
                                      % self.endpoint)

        result = json.loads(data)

        self.crumb = result['crumb']
        self.crumb_request_field = result['crumbRequestField']

    def _make_request(self, url, body=None, method='GET',
                      content_type=''):
        """Make an HTTP request.

        This will first attempt to fetch a CSRF token if we do not currently
        have one.

        Args:
            url (unicode):
                The URL to make the request against.

            body (unicode or bytes, optional):
                The content of the request.

            method (unicode, optional):
                The request method. If not provided, it defaults to a ``GET``
                request.

            content_type (unicode, optional):
                The type of the content being POSTed.

        Returns:
            bytes:
            The contents of the HTTP response body.

        Raises:
            urllib2.URLError:
                The HTTP request failed.
        """
        if self.csrf_protection_enabled and not self.crumb:
            try:
                self._fetch_csrf_token()
            except HTTPError as e:
                if e.code == 404:
                    self.csrf_protection_enabled = False
                else:
                    raise e

        return self._make_raw_request(url, body, method, content_type)

    def _make_raw_request(self, url, body=None, method='GET',
                          content_type=''):
        """Make an HTTP request.

        Args:
            url (unicode):
                The URL to make the request against.

            body (unicode or bytes, optional):
                The content of the request.

            method (unicode, optional):
                The request method. If not provided, it defaults to a ``GET``
                request.

            content_type (unicode, optional):
                The type of the content being POSTed.

        Returns:
            bytes:
            The contents of the HTTP response body.

        Raises:
            urllib2.URLError:
                The HTTP request failed.
        """
        logger.debug('Making request to Jenkins CI %s', url)

        headers = {}

        if self.crumb:
            headers[self.crumb_request_field] = self.crumb

        if content_type:
            headers['Content-Type'] = content_type

        if isinstance(body, str):
            body = body.encode('utf-8')

        request = HostingServiceHTTPRequest(
            url,
            body=body,
            method=method,
            headers=headers)
        request.add_basic_auth(self.username, self.password)

        return self._open_request(request)

    def _open_request(self, request):
        """Perform an HTTP request.

        Args:
            request (reviewboard.hostingsvcs.service.
                     HostingServiceHTTPRequest):
                The HTTP request object.

        Returns:
            bytes:
            The response data.
        """
        response = request.open()
        return response.data



"""Forms for chat integrations."""

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm

from rbintegrations.util.conditions import ReviewRequestConditionsField


class BaseChatIntegrationConfigForm(IntegrationConfigForm):
    """Form for configuring chat integration.

    This allows an administrator to set up a chat configuration for sending
    messages to a given WebHook URL based on the specified
    conditions.
    """

    conditions = ReviewRequestConditionsField()

    webhook_url = forms.CharField(
        label=_('Webhook URL'),
        required=True,
        help_text=_('Your unique webhook URL. This can be '
                    'found in the "Setup Instructions" box inside the '
                    'Incoming WebHooks integration.'),
        widget=forms.TextInput(attrs={
            'size': 80,
        }))

    channel = forms.CharField(
        label=_('Send to Channel'),
        required=False,
        help_text=_('The optional name of the channel review request updates '
                    'are sent to. By default, the configured channel on the '
                    'Incoming Webhook will be used.'),
        widget=forms.TextInput(attrs={
            'size': 40,
        }))

    class Meta:
        fieldsets = (
            ('What To Post', {
                'description': _(
                    'You can choose which review requests would be posted by '
                    'choosing the repositories and groups to match against.'
                ),
                'fields': ('conditions',),
            }),
            ('Where To Post', {
                'description': _(
                    'To start, add a new "Incoming WebHooks" service '
                    'integration. You can then provide the "Unique WebHook '
                    'URL" below, and optionally choose a custom channel to '
                    'send notifications to.'
                ),
                'fields': ('webhook_url', 'channel'),
                'classes': ('wide',)
            }),
        )



"""Integration for building changes on CircleCI."""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING
from urllib.parse import quote_plus
from urllib.request import urlopen

from django.utils.functional import cached_property

from rbintegrations.baseci.errors import CIBuildError
from rbintegrations.baseci.integration import (BaseCIIntegration,
                                               BuildPrepData)
from rbintegrations.circleci.forms import CircleCIIntegrationConfigForm
from rbintegrations.util.compat.logs import log_timed
from rbintegrations.util.urlrequest import URLRequest

if TYPE_CHECKING:
    from reviewboard.integrations.models import IntegrationConfig
    from reviewboard.reviews.models.status_update import StatusUpdate


logger = logging.getLogger(__name__)


class CircleCIIntegration(BaseCIIntegration):
    """Integrates Review Board with CircleCI."""

    name = 'CircleCI'
    description = 'Builds diffs posted to Review Board using CircleCI.'
    config_form_cls = CircleCIIntegrationConfigForm

    status_update_service_id = 'circle-ci'

    bot_username = 'circle-ci'
    bot_user_first_name = 'Circle'
    bot_user_last_name = 'CI'

    @cached_property
    def icon_static_urls(self):
        """Return the icons used for the integration.

        Returns:
            dict:
            The icons for CircleCI.
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        return {
            '1x': extension.get_static_url('images/circleci/icon.png'),
            '2x': extension.get_static_url('images/circleci/icon@2x.png'),
        }

    def prepare_builds(
        self,
        prep_data: BuildPrepData,
    ) -> bool:
        """Prepare for builds.

        This will check if the change is on a supported hosting service
        (GitHub or Bitbucket), and only permit the build if it is.

        Version Added:
            3.1

        Args:
            prep_data (BuildPrepData):
                The builds preparation data to modify.

        Returns:
            bool:
            ``True`` if the builds should proceed. ``False`` if builds should
            be skipped.
        """
        # Only build changes against GitHub or Bitbucket repositories.
        repository = prep_data.repository

        if not repository or not repository.hosting_account:
            return False

        service_name = repository.hosting_account.service_name

        # This may look weird, but it's here for defensive purposes.
        # Currently, the possible values for CircleCI's "vcs-type" field in
        # their API happens to match up perfectly with our service names,
        # but that's not necessarily always going to be the case.
        vcs_type: str

        if service_name == 'github':
            vcs_type = 'github'
        elif service_name == 'bitbucket':
            vcs_type = 'bitbucket'
        else:
            return False

        prep_data.extra_state['vcs_type'] = vcs_type

        return True

    def start_build(
        self,
        *,
        prep_data: BuildPrepData,
        config: IntegrationConfig,
        status_update: StatusUpdate,
    ) -> None:
        """Start a new build.

        This will trigger a CircleCI build for the given review request and
        configuration.

        Version Added:
            3.1

        Args:
            prep_data (BuildPrepData):
                The builds preparation data containing information for the
                build.

            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration for the integration triggering this build.

            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update indicating the status of this build.

        Raises:
            rbintegrations.baseci.errors.CIBuildError:
                There was an error invoking the build.
        """
        vcs_type = prep_data.extra_state['vcs_type']
        diffset = prep_data.diffset
        local_site = prep_data.local_site
        repository = prep_data.repository
        review_request = prep_data.review_request

        org_name, repo_name = self._get_repo_ids(vcs_type, repository)
        circleci_api_token = config.get('circle_api_token')

        if not circleci_api_token:
            logger.error('Unable to make CircleCI API request for '
                         'integration config %d: api_token is missing.',
                         config.pk)
            raise CIBuildError('missing API token.')

        url = ('https://circleci.com/api/v1.1/project/%s/%s/%s/tree/%s'
               '?circle-token=%s'
               % (vcs_type, org_name, repo_name,
                  config.get('branch_name') or 'master',
                  quote_plus(circleci_api_token.encode('utf-8'))))

        api_token = self.get_or_create_api_token(user=prep_data.user,
                                                 local_site=local_site)

        body = {
            'revision': diffset.base_commit_id,
            'build_parameters': {
                'CIRCLE_JOB': 'reviewboard',
                'REVIEWBOARD_SERVER': prep_data.server_url,
                'REVIEWBOARD_REVIEW_REQUEST': review_request.display_id,
                'REVIEWBOARD_DIFF_REVISION': diffset.revision,
                'REVIEWBOARD_API_TOKEN': api_token.token,
                'REVIEWBOARD_STATUS_UPDATE_ID': status_update.pk,
            },
        }

        if local_site:
            body['build_parameters']['REVIEWBOARD_LOCAL_SITE'] = \
                local_site.name

        request = URLRequest(
            url,
            body=json.dumps(body),
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
            },
            method='POST')

        with log_timed(f'Sending build request to CircleCI API at {url}',
                       logger=logger) as log_timer:
            try:
                u = urlopen(request)
                data = json.loads(u.read())

                self.update_status(status_update,
                                   url=data['build_url'],
                                   url_text='View Build')
            except Exception as e:
                logger.exception('[%s] Failed to start build: %s',
                                 log_timer.trace_id, e)

    def _get_repo_ids(self, service_name, repository):
        """Return the organization and repo name for the given repository.

        Args:
            service_name (unicode):
                The version control service name.

            repository (reviewboard.scmtools.models.Repository):
                The repository.

        Returns:
            tuple of unicode:
            A two-tuple consisting of the organization (or user) and repository
            names.
        """
        extra_data = repository.extra_data
        plan = extra_data['repository_plan']

        if service_name == 'github':
            if plan == 'public':
                return (repository.hosting_account.username,
                        extra_data['github_public_repo_name'])
            elif plan == 'public-org':
                return (extra_data['github_public_org_name'],
                        extra_data['github_public_org_repo_name'])
            elif plan == 'private':
                return (repository.hosting_account.username,
                        extra_data['github_private_repo_name'])
            elif plan == 'private-org':
                return (extra_data['github_private_org_name'],
                        extra_data['github_private_org_repo_name'])
            else:
                raise ValueError('Unexpected plan for GitHub repository %d: %s'
                                 % (repository.pk, plan))
        elif service_name == 'bitbucket':
            if plan == 'personal':
                return (extra_data['bitbucket_account_username'],
                        extra_data['bitbucket_repo_name'])
            elif plan == 'other-user':
                return (extra_data['bitbucket_other_user_username'],
                        extra_data['bitbucket_other_user_repo_name'])
            elif plan == 'team':
                return (extra_data['bitbucket_team_name'],
                        extra_data['bitbucket_team_repo_name'])
            else:
                raise ValueError('Unexpected plan for Bitbucket repository '
                                 '%d: %s'
                                 % (repository.pk, plan))
        else:
            raise ValueError('Unexpected service_name %s' % service_name)



"""Forms for I Done This integration."""

import logging
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

from django import forms
from django.utils.translation import gettext, gettext_lazy as _
from reviewboard.accounts.forms.pages import AccountPageForm
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.scmtools.crypto_utils import encrypt_password

from rbintegrations.idonethis.utils import (create_idonethis_request,
                                            delete_cached_user_team_ids,
                                            get_user_api_token)
from rbintegrations.util.conditions import ReviewRequestConditionsField


class IDoneThisIntegrationConfigForm(IntegrationConfigForm):
    """Admin configuration form for I Done This.

    This allows an administrator to set up a configuration for posting 'done'
    entries to a given I Done This team based on the specified conditions.
    """

    conditions = ReviewRequestConditionsField()

    team_id = forms.CharField(
        label=_('Team ID'),
        required=True,
        help_text=_('The identifier of the team to receive posts. This can '
                    'be found at the end of the team URL, e.g. '
                    '<code>https://beta.idonethis.com/t/'
                    '<strong>123456abcdef</strong></code>'),
        widget=forms.TextInput(attrs={
            'size': 15,
        }))

    def clean_team_id(self):
        """Clean and validate the 'team_id' field.

        Returns:
            unicode:
            Team ID with leading and trailing whitespace removed.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if the team ID contains any slashes.
        """
        team_id = self.cleaned_data['team_id'].strip()

        if '/' in team_id:
            raise forms.ValidationError(
                gettext('Team ID cannot contain slashes.'))

        return team_id

    class Meta:
        fieldsets = (
            (_('What to Post'), {
                'description': _(
                    'You can choose which review request activity would be '
                    'posted by selecting the conditions to match.'
                ),
                'fields': ('conditions',),
            }),
            (_('Where to Post'), {
                'description': _(
                    'Posts are made to the specified I Done This team on '
                    'behalf of individual users who belong to that team. '
                    'A separate configuration is required for each team, '
                    'and multiple configurations may use the same team to '
                    'specify alternative sets of conditions.\n'
                    'To enable posting, each user has to provide their '
                    'personal I Done This API Token on their Review Board '
                    'account page.'
                ),
                'fields': ('team_id',),
                'classes': ('wide',)
            }),
        )


class IDoneThisIntegrationAccountPageForm(AccountPageForm):
    """User account page form for I Done This.

    This allows a user to specify their I Done This API Token and choose
    which actions result in posting of 'done' entries to I Done This.
    """

    form_id = 'idonethis_account_page_form'
    form_title = _('I Done This')
    template_name = 'rbintegrations/idonethis/account_page_form.html'

    idonethis_api_token = forms.CharField(
        label=_('API Token'),
        required=False,
        widget=forms.TextInput(attrs={
            'size': 45,
        }))

    def clean_idonethis_api_token(self):
        """Clean and validate the 'idonethis_api_token' field.

        This performs a test against the I Done This authentication test
        endpoint to ensure that the provided API token is valid. We only care
        if the request is successful, so we ignore the returned user data.

        Returns:
            unicode:
            Validated API token with leading and trailing whitespace removed,
            or an empty string if the API token is empty.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if the API token validation fails.
        """
        api_token = self.cleaned_data['idonethis_api_token'].strip()

        if not api_token:
            return ''

        request = create_idonethis_request('noop', api_token)
        logging.debug('IDoneThis: Validating API token for user "%s", '
                      'request "%s %s"',
                      self.user.username,
                      request.get_method(),
                      request.get_full_url())

        try:
            urlopen(request)
        except (HTTPError, URLError) as e:
            if isinstance(e, HTTPError):
                error_info = '%s, error data: %s' % (e, e.read())
            else:
                error_info = e.reason

            logging.error('IDoneThis: Failed to validate API token for user '
                          '"%s", request "%s %s": %s',
                          self.user.username,
                          request.get_method(),
                          request.get_full_url(),
                          error_info)

            raise forms.ValidationError(
                gettext('Error validating the API Token. Make sure the token '
                        'matches your I Done This Account Settings.'))

        return api_token

    def load(self):
        """Load the account page form."""
        self.set_initial({
            'idonethis_api_token': get_user_api_token(self.user),
        })

    def save(self):
        """Save the account page form.

        Stores an encrypted version of the API token.
        """
        api_token = self.cleaned_data['idonethis_api_token']
        profile = self.user.get_profile()
        settings = profile.settings.setdefault('idonethis', {})

        if api_token:
            logging.debug('IDoneThis: Saving API token for user "%s"',
                          self.user.username)
            settings['api_token'] = encrypt_password(api_token)
        elif 'api_token' in settings:
            logging.debug('IDoneThis: Deleting API token for user "%s"',
                          self.user.username)
            del settings['api_token']

        profile.save()

        delete_cached_user_team_ids(self.user)



"""Integration for Slack."""

from __future__ import annotations

import json
import logging
from typing import (MutableMapping, Optional, Sequence, TYPE_CHECKING, Union,
                    cast)
from urllib.request import Request, urlopen

from django.utils.functional import cached_property
from djblets.util.typing import JSONListImmutable
from housekeeping.functions import deprecate_non_keyword_only_args
from reviewboard.admin.server import build_server_url

from rbintegrations.basechat.integration import BaseChatIntegration
from rbintegrations.deprecation import RemovedInRBIntegrations50Warning
from rbintegrations.util.compat.logs import log_timed

if TYPE_CHECKING:
    from djblets.util.typing import JSONDict, JSONValue
    from rbintegrations.basechat.integration import FieldsDict
    from reviewboard.reviews.models import ReviewRequest
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


@deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
def build_slack_message(
    *,
    integration: BaseChatIntegration,
    title: str,
    title_link: str,
    fallback_text: str,
    fields: Sequence[FieldsDict] = [],
    pre_text: Optional[str] = None,
    body: Optional[str] = None,
    color: Optional[str] = None,
    thumb_url: Optional[str] = None,
    image_url: Optional[str] = None,
) -> JSONDict:
    """Build message using Slack WebHook format.

    This will build the payload data for HTTP requests to services such as
    Slack, Mattermost and Discord.

    Version Changed:
        4.0:
        * Made all arguments keyword-only.

    Args:
        integration (BaseChatIntegration):
            The Integration.

        title (str):
            The title for the message.

        title_link (str):
            The link for the title of the message.

        fallback_text (str):
            The non-rich fallback text to display in the chat, for use in
            IRC and other services.

        fields (list of rbintegrations.basechat.integration.FieldsDict,
                optional):
            The fields comprising the rich message to display in chat.

        pre_text (str, optional):
            Text to display before the rest of the message.

        body (str, optional):
            The body of the message.

        color (str, optional):
            A Slack color string or RGB hex value for the message.

        thumb_url (str, optional):
            URL of an image to show on the side of the message.

        image_url (str, optional):
            URL of an image to show in the message.

    Returns:
        djblets.util.typing.JSONDict:
        The payload of the Slack message request.
    """
    if not color:
        color = integration.default_color

    attachment: dict[str, Union[JSONValue, Sequence[FieldsDict]]] = {
        'color': color or integration.default_color,
        'fallback': fallback_text,
        'title': title,
        'title_link': title_link,
        'text': body,
        'pretext': pre_text,
    }

    if fields:
        attachment['fields'] = fields

    if thumb_url:
        attachment['thumb_url'] = thumb_url

    if image_url:
        attachment['image_url'] = image_url

    return {
        'attachments': cast(JSONListImmutable, [attachment]),
        'icon_url': integration.logo_url,
    }


@deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
def notify(
    *,
    integration: BaseChatIntegration,
    title: str,
    title_link: str,
    fallback_text: str,
    local_site: Optional[LocalSite],
    review_request: ReviewRequest,
    event_name: Optional[str] = None,
    fields: Sequence[FieldsDict] = [],
    pre_text: Optional[str] = None,
    body: Optional[str] = None,
    color: Optional[str] = None,
    thumb_url: Optional[str] = None,
    image_url: Optional[str] = None,
) -> None:
    """Send a WebHook notification.

    This will post the given message to any Slacks/Mattermost channels
    configured to receive it. This is oriented towards Slack, however is
    broken out of the SlackIntegration because other services (like
    Mattermost) duplicate Slack APIs.

    Version Changed:
        4.0:
        * Made all arguments keyword-only.

    Args:
        integration (BaseChatIntegration):
            The Integration.

        title (str):
            The title for the message.

        title_link (str):
            The link for the title of the message.

        fallback_text (str):
            The non-rich fallback text to display in the chat, for use in
            IRC and other services.

        local_site (reviewboard.site.models.LocalSite):
            The Local Site for the review request or review emitting
            the message. Only integration configurations matching this
            Local Site will be processed.

        review_request (reviewboard.reviews.models.ReviewRequest):
            The review request the notification is bound to.

        event_name (str):
            The name of the event triggering this notification.

        fields (list of rbintegrations.basechat.integration.FieldsDict,
                optional):
            The fields comprising the rich message to display in chat.

        pre_text (str, optional):
            Text to display before the rest of the message.

        body (str, optional):
            The body of the message.

        color (str, optional):
            A Slack color string or RGB hex value for the message.

        thumb_url (str, optional):
            URL of an image to show on the side of the message.

        image_url (str, optional):
            URL of an image to show in the message.
    """
    common_payload = build_slack_message(integration=integration,
                                         title=title,
                                         title_link=title_link,
                                         fallback_text=fallback_text,
                                         fields=fields,
                                         pre_text=pre_text,
                                         body=body,
                                         color=color,
                                         thumb_url=thumb_url,
                                         image_url=image_url)

    # Send a notification to any configured channels.
    for config in integration.get_configs(local_site):
        if not config.match_conditions(form_cls=integration.config_form_cls,
                                       review_request=review_request):
            continue

        payload = dict({
            'username': config.get('notify_username'),
        }, **common_payload)

        channel = config.get('channel')

        if channel:
            payload['channel'] = channel

        webhook_url = config.get('webhook_url')

        with log_timed(f'Sending notification for event "{event_name}", '
                       f'review_request ID {review_request.pk} to channel '
                       f'"{channel}", WebHook URL {webhook_url}',
                       logger=logger) as log_timer:
            try:
                if not webhook_url:
                    raise Exception('WebHook URL has not been configured.')

                data = json.dumps(payload).encode('utf-8')
                headers: MutableMapping[str, str] = {
                    'Content-Type': 'application/json',
                    'Content-Length': str(len(data)),
                }
                urlopen(Request(webhook_url, data, headers))
            except Exception as e:
                logger.exception('[%s] Failed to send notification: %s',
                                 log_timer.trace_id, e)


@deprecate_non_keyword_only_args(RemovedInRBIntegrations50Warning)
def format_link(
    *,
    path: str,
    text: str,
) -> str:
    """Format the given URL and text to be shown in a message.

    This will combine together the parts of the URL (method, domain, path)
    and format it using Slack/Mattermost's URL syntax. This is oriented
    towards Slack, however is broken out of the SlackIntegration because
    other services (like Mattermost) duplicate Slack APIs.

    Version Changed:
        4.0:
        * Made all arguments keyword-only.

    Args:
        path (str):
            The path on the Review Board server.

        text (str):
            The text for the link.

    Returns:
        str:
        The link for use in Slack.
    """
    # Slack/Mattermost only want these three entities replaced, rather than
    # all the entities that Django's escape() would attempt to replace.
    text = text.replace('&', '&amp;')
    text = text.replace('<', '&lt;')
    text = text.replace('>', '&gt;')

    return '<%s|%s>' % (build_server_url(path), text)


class SlackIntegration(BaseChatIntegration):
    """Integrates Review Board with Slack.

    This will handle updating Slack channels when review requests are posted,
    changed, or closed, and when there's new activity on the review request.
    """

    name = 'Slack'
    description = (
        'Notifies channels in Slack when review requests are created, '
        'updated, and reviewed.'
    )

    default_settings = {
        'webhook_url': '',
        'channel': '',
        'notify_username': 'Review Board',
    }

    assets_base_url = (
        'https://static.reviewboard.org/integration-assets/slack'
    )

    def notify(
        self,
        *,
        title: str,
        title_link: str,
        fallback_text: str,
        local_site: Optional[LocalSite],
        review_request: ReviewRequest,
        event_name: Optional[str] = None,
        fields: Sequence[FieldsDict] = [],
        pre_text: Optional[str] = None,
        body: Optional[str] = None,
        color: Optional[str] = None,
        thumb_url: Optional[str] = None,
        image_url: Optional[str] = None,
    ) -> None:
        """Send a WebHook notification to Slack.

        This will post the given message to any Slack channels configured to
        receive it.

        Args:
            title (str):
                The title for the message.

            title_link (str):
                The link for the title of the message.

            fallback_text (str):
                The non-rich fallback text to display in the chat, for use in
                IRC and other services.

            fields (list of rbintegrations.basechat.integration.FieldsDict,
                    optional):
                The fields comprising the rich message to display in chat.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site for the review request or review emitting
                the message. Only integration configurations matching this
                Local Site will be processed.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request the notification is bound to.

            event_name (str):
                The name of the event triggering this notification.

            pre_text (str, optional):
                Text to display before the rest of the message.

            body (str, optional):
                The body of the message.

            color (str, optional):
                A Slack color string or RGB hex value for the message.

            thumb_url (str, optional):
                URL of an image to show on the side of the message.

            image_url (str, optional):
                URL of an image to show in the message.
        """
        notify(integration=self,
               title=title,
               title_link=title_link,
               fallback_text=fallback_text,
               local_site=local_site,
               review_request=review_request,
               event_name=event_name,
               fields=fields,
               pre_text=pre_text,
               body=body,
               color=color,
               thumb_url=thumb_url,
               image_url=image_url)

    def format_link(
        self,
        *,
        path: str,
        text: str,
    ) -> str:
        """Format the given URL and text to be shown in a Slack message.

        This will combine together the parts of the URL (method, domain, path)
        and format it using Slack's URL syntax.

        Args:
            path (str):
                The path on the Review Board server.

            text (str):
                The text for the link.

        Returns:
            str:
            The link for use in Slack.
        """
        return format_link(path=path, text=text)

    # NOTE: It's technically type-unsafe to replace an attribute with a
    #       property, but this works for us practically.
    @cached_property
    def icon_static_urls(self) -> dict[str, str]:  # type: ignore
        """Return the icons used for the integration.

        Returns:
            dict:
            The icons for Slack.
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        assert extension is not None

        return {
            '1x': extension.get_static_url('images/slack/icon.png'),
            '2x': extension.get_static_url('images/slack/icon@2x.png'),
        }



"""Views for the Trello integration."""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING
from urllib.parse import urlencode
from urllib.request import urlopen

from django.http import HttpResponse
from django.views.generic import View
from reviewboard.integrations.base import get_integration_manager
from reviewboard.reviews.views import ReviewRequestViewMixin

from rbintegrations.util.compat.logs import log_timed

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class TrelloCardSearchView(ReviewRequestViewMixin, View):
    """The view to search for Trello cards (for use with auto-complete)."""

    def get(
        self,
        request: HttpRequest,
        **kwargs,
    ) -> HttpResponse:
        """Perform a search for cards.

        Args:
            request (django.http.HttpRequest):
                The request.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.http.HttpResponse:
            A response containing JSON with a list of cards matching the
            search criteria.
        """
        from rbintegrations.trello.integration import TrelloIntegration

        integration_manager = get_integration_manager()
        integration = integration_manager.get_integration(
            TrelloIntegration.integration_id)

        configs = (
            config
            for config in integration.get_configs(
                self.review_request.local_site)
            if config.match_conditions(form_cls=integration.config_form_cls,
                                       review_request=self.review_request)
        )

        results = []

        query = request.GET.get('q')

        if not query:
            return HttpResponse(json.dumps([]),
                                content_type='application/json')

        params = {
            'card_board': 'true',
            'card_fields': 'id,name,shortUrl',
            'card_list': 'true',
            'cards_limit': 20,
            'modelTypes': 'cards',
            'partial': 'true',
            'query': query,
        }

        for config in configs:
            params['key'] = config.settings['trello_api_key']
            params['token'] = config.settings['trello_api_token']

            url = 'https://api.trello.com/1/search?%s' % urlencode(params)

            with log_timed(f'Fetching Trello cards at {url}',
                           logger=logger,
                           request=request) as log_timer:
                try:
                    response = urlopen(url)
                    data = json.loads(response.read())

                    for card in data['cards']:
                        results.append({
                            'id': card['id'],
                            'name': card['name'],
                            'board': card.get('board', {}).get('name', ''),
                            'list': card.get('list', {}).get('name', ''),
                            'url': card['shortUrl'],
                        })
                except Exception as e:
                    logger.exception('[%s] Unexpected error when searching '
                                     'for Trello cards: %s',
                                     log_timer.trace_id, e)

        return HttpResponse(json.dumps(results),
                            content_type='application/json')



"""Review Request field definitions for Trello integration."""

from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.base import get_integration_manager
from reviewboard.reviews.fields import BaseEditableField


class TrelloField(BaseEditableField):
    """A review request field for selecting Trello cards."""

    field_id = 'rbintegrations_trello'
    label = _('Trello Cards')

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.TrelloFieldView'

    #: The HTML tag to be used when rendering the field.
    tag_name = 'div'

    @property
    def should_render(self):
        """Whether the field should render or not.

        The Trello Cards field renders when there are any matching
        integration configs for the current review request.
        """
        from rbintegrations.trello.integration import TrelloIntegration

        integration_manager = get_integration_manager()
        integration = integration_manager.get_integration(
            TrelloIntegration.integration_id)
        review_request = self.review_request_details.get_review_request()

        for config in integration.get_configs(review_request.local_site):
            if config.match_conditions(form_cls=integration.config_form_cls,
                                       review_request=review_request):
                return True

        return False

    def render_value(self, value):
        """Render the value in the field.

        This always returns the empty string, because the field value is
        rendered in JavaScript.

        Args:
            value (object):
                The value to render.

        Returns:
            unicode:
            The rendered value.
        """
        return ''

    def get_data_attributes(self):
        """Return any data attributes to include in the element.

        Returns:
            dict:
            The data attributes to include in the element.
        """
        attrs = super(TrelloField, self).get_data_attributes()

        if self.value is not None:
            attrs['raw-value'] = self.value

        return attrs



"""The form for configuring the Jenkins CI integration."""

import logging
from urllib.error import HTTPError, URLError

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.reviews.conditions import ReviewRequestConditionChoices

from rbintegrations.baseci.forms import BaseCIIntegrationConfigForm
from rbintegrations.jenkinsci.api import JenkinsAPI


logger = logging.getLogger(__name__)


class JenkinsCIIntegrationConfigForm(BaseCIIntegrationConfigForm):
    """Form for configuring Jenkins CI."""

    jenkins_endpoint = forms.URLField(
        label=_('Server'),
        help_text=_('Server endpoint URL.'))

    jenkins_job_name = forms.CharField(
        label=_('Job Name'),
        help_text=_(
            'Job name. This can include the following variables: '
            '{branch_name} (the branch name), '
            '{repository_name} (the repository name), '
            '{noslash_branch_name} (the branch name with slashes converted to '
            'underscores), and {noslash_repository_name} (the repository name '
            'with slashes converted to underscores).'
        ))

    jenkins_username = forms.CharField(
        label=_('Username'),
        help_text=_('User who has access to the above job.'))

    jenkins_password = forms.CharField(
        label=_('API Token / Password'),
        help_text=_('The API token used for authentication. Older versions '
                    'may require a user password instead.'),
        widget=forms.PasswordInput)

    jenkins_user_token = forms.CharField(
        label=_('Review Board API Token'),
        help_text=_('This API token is used by Jenkins to update build '
                    'status. Please specify this in the Jenkins-side '
                    'configuration. Note that if you switch the local site '
                    'this API token will be updated upon saving.'),
        required=False,
        widget=forms.TextInput(attrs={'readonly': True}))

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the form.

        Args:
            *args (tuple):
                Arguments for the form.

            **kwargs (dict):
                Keyword arguments for the form.
        """
        super().__init__(*args, **kwargs)

        integration = self.integration
        local_site = self.fields['local_site'].initial
        user = integration.get_or_create_user()

        token = integration.get_or_create_api_token(
            user=user,
            local_site=local_site)

        self.initial['jenkins_user_token'] = token.token

    def clean(self):
        """Clean the form.

        This validates the user credentials that have been provided.

        Returns:
            dict:
            The cleaned data.
        """
        cleaned_data = super(JenkinsCIIntegrationConfigForm, self).clean()

        if self._errors:
            # Form validation has already failed.
            return cleaned_data

        api = JenkinsAPI(endpoint=cleaned_data.get('jenkins_endpoint'),
                         job_name=cleaned_data.get('jenkins_job_name'),
                         username=cleaned_data.get('jenkins_username'),
                         password=cleaned_data.get('jenkins_password'))

        try:
            # Tests a simple endpoint to ensure the user credentials are
            # correct.
            api.test_connection()
        except HTTPError as e:
            if e.code == 403:
                message = _('Unable to authenticate with the provided user.')
            elif e.code == 401:
                message = _('Provided user credentials are incorrect.')
            else:
                message = str(e)

            self._errors['jenkins_username'] = self.error_class([message])
            self._errors['jenkins_password'] = self.error_class([message])

            return cleaned_data
        except URLError as e:
            self._errors['jenkins_endpoint'] = self.error_class([e])
            return cleaned_data

        return cleaned_data

    class Meta(BaseCIIntegrationConfigForm.Meta):
        fieldsets = (
            (_('What To Build'), {
                'description': _(
                    'You can choose which review requests to build using this '
                    'configuration by setting conditions here. At a minimum, '
                    'this should include the specific repository to use '
                    'this configuration for.'
                ),
                'fields': (
                    'conditions',
                ),
            }),
            (_('Where To Build'), {
                'description': _(
                    'Configure the address and authentication credentials '
                    'for the Jenkins server handling your builds.'
                ),
                'fields': (
                    'jenkins_endpoint',
                    'jenkins_username',
                    'jenkins_password',
                ),
            }),
            (_('How To Build'), {
                'description': _(
                    "Builds performed on the code in review requests will use "
                    "a completely separate configuration from commits which "
                    "are pushed to the repository. The configuration listed "
                    "here will be used regardless of and in addition to any "
                    "external configurations."
                    "\n"
                    "It's also recommended to create a special branch head "
                    "in the repository to use for these builds, so they "
                    "don't appear to be happening on your main development "
                    "branch. This branch can contain anything (or even be "
                    "empty), since the code will come from the review "
                    "request."
                ),
                'fields': (
                    'jenkins_job_name',
                    'jenkins_user_token',
                ),
            }),
            (_('When To Build'), {
                'fields': (
                    'run_manually',
                    'timeout',
                ),
            }),
        )



"""Utilities for interacting with the Travis CI API."""

import logging
import json
from urllib.parse import quote_plus
from urllib.request import urlopen

from django.utils.translation import gettext_lazy as _

from rbintegrations.util.urlrequest import URLRequest


logger = logging.getLogger(__name__)


class TravisAPI(object):
    """Object for interacting with the Travis CI API."""

    OPEN_SOURCE_ENDPOINT = 'O'
    PRIVATE_PROJECT_ENDPOINT = 'P'
    ENTERPRISE_ENDPOINT = 'E'

    ENDPOINT_CHOICES = (
        (OPEN_SOURCE_ENDPOINT, _('Open Source (travis-ci.org)')),
        (PRIVATE_PROJECT_ENDPOINT, _('Private Projects (travis-ci.com)')),
        (ENTERPRISE_ENDPOINT, _('Enterprise (custom domain)')),
    )

    OPEN_SOURCE_ENDPOINT_URL = 'https://api.travis-ci.org'
    PRIVATE_PROJECT_ENDPOINT_URL = 'https://api.travis-ci.com'

    def __init__(self, config):
        """Initialize the object.

        Args:
            config (dict):
                The integration config to use.

        Raises:
            ValueError:
                The provided endpoint type was not valid.
        """
        endpoint = config.get('travis_endpoint')

        if endpoint == self.OPEN_SOURCE_ENDPOINT:
            self.endpoint = self.OPEN_SOURCE_ENDPOINT_URL
        elif endpoint == self.PRIVATE_PROJECT_ENDPOINT:
            self.endpoint = self.PRIVATE_PROJECT_ENDPOINT_URL
        elif endpoint == self.ENTERPRISE_ENDPOINT:
            custom_endpoint = config.get('travis_custom_endpoint')

            if custom_endpoint.endswith('/'):
                custom_endpoint = custom_endpoint[:-1]

            self.endpoint = '%s/api' % custom_endpoint
        else:
            raise ValueError('Unexpected value for Travis CI endpoint: %s'
                             % endpoint)

        self.token = config.get('travis_ci_token')

    def lint(self, travis_yml):
        """Lint a prospective travis.yml file.

        Args:
            travis_yml (unicode):
                The contents of the travis.yml file to validate.

        Returns:
            dict:
            The parsed contents of the JSON response.

        Raises:
            urllib2.URLError:
                The HTTP request failed.

            Exception:
                Some other exception occurred when trying to parse the results.
        """
        data = self._make_request('%s/lint' % self.endpoint,
                                  body=travis_yml,
                                  method='POST',
                                  content_type='text/yaml')
        return json.loads(data)

    def get_config(self):
        """Return the Travis CI server's config.

        Returns:
            dict:
            The parsed contents of the JSON response.

        Raises:
            urllib2.URLError:
                The HTTP request failed.
        """
        # This request can't go through _make_request because this endpoint
        # isn't available with API version 3 and doesn't require
        # authentication.
        u = urlopen(URLRequest('%s/config' % self.endpoint))
        return json.loads(u.read())

    def get_user(self):
        """Return the Travis CI user.

        Returns:
            dict:
            The parsed contents of the JSON response.

        Raises:
            urllib2.URLError:
                The HTTP request failed.
        """
        data = self._make_request('%s/user' % self.endpoint)
        return json.loads(data)

    def start_build(self, repo_slug, travis_config, commit_message,
                    branch=None):
        """Start a build.

        Args:
            repo_slug (unicode):
                The "slug" for the repository based on it's location on GitHub.

            travis_config (unicode):
                The contents of the travis config to use when doing the build.

            commit_message (unicode):
                The text to use as the commit message displayed in the Travis
                UI.

            branch (unicode, optional):
                The branch name to use.

        Returns:
            dict:
            The parsed contents of the JSON response.

        Raises:
            urllib2.URLError:
                The HTTP request failed.
        """
        travis_config['merge_mode'] = 'replace'

        request_data = {
            'request': {
                'message': commit_message,
                'config': travis_config,
            },
        }

        if branch:
            request_data['request']['branch'] = branch

        data = self._make_request(
            '%s/repo/%s/requests' % (self.endpoint,
                                     quote_plus(repo_slug)),
            body=json.dumps(request_data),
            method='POST',
            content_type='application/json')

        return json.loads(data)

    def _make_request(self, url, body=None, method='GET',
                      content_type='application/json'):
        """Make an HTTP request.

        Args:
            url (unicode):
                The URL to make the request against.

            body (unicode or bytes, optional):
                The content of the request.

            method (unicode, optional):
                The request method. If not provided, it defaults to a ``GET``
                request.

            content_type (unicode, optional):
                The type of the content being POSTed.

        Returns:
            bytes:
            The contents of the HTTP response body.

        Raises:
            urllib2.URLError:
                The HTTP request failed.
        """
        logger.debug('Making request to Travis CI %s', url)

        headers = {
            'Accept': 'application/json',
            'Authorization': 'token %s' % self.token,
            'Travis-API-Version': '3',
        }

        if content_type:
            headers['Content-Type'] = content_type

        request = URLRequest(
            url,
            body=body,
            method=method,
            headers=headers)

        u = urlopen(request)
        return u.read()



"""Unit tests for the CircleCI integration."""

import json
from urllib.request import urlopen

from djblets.conditions import ConditionSet, Condition
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.hostingsvcs.service import (HostingServiceHTTPRequest,
                                             HostingServiceHTTPResponse)
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.signals import status_update_request_run

from rbintegrations.circleci.integration import CircleCIIntegration
from rbintegrations.testing.testcases import IntegrationTestCase


class CircleCIIntegrationTests(IntegrationTestCase):
    """Tests for CircleCI."""

    integration_cls = CircleCIIntegration
    fixtures = ['test_scmtools', 'test_site', 'test_users']

    def test_build_new_review_request(self):
        """Testing CircleCIIntegration builds a new review request"""
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_urlopen()

        review_request.publish(review_request.submitter)

        self.assertTrue(urlopen.spy.called)
        self.assertEqual(data['url'],
                         'https://circleci.com/api/v1.1/project/github/'
                         'mypublicorg/mypublicorgrepo/tree/review-requests?'
                         'circle-token=test-api-token')

        self.assertEqual(data['build_params']['CIRCLE_JOB'], 'reviewboard')
        self.assertIn('REVIEWBOARD_API_TOKEN', data['build_params'])
        self.assertIn('REVIEWBOARD_DIFF_REVISION', data['build_params'])
        self.assertIn('REVIEWBOARD_REVIEW_REQUEST', data['build_params'])
        self.assertIn('REVIEWBOARD_SERVER', data['build_params'])
        self.assertIn('REVIEWBOARD_STATUS_UPDATE_ID', data['build_params'])

    def test_build_new_review_request_with_local_site(self):
        """Testing CircleCIIntegration builds a new review request with a local
        site
        """
        repository = self._create_repository(with_local_site=True)
        review_request = self.create_review_request(repository=repository,
                                                    with_local_site=True)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        data = self._spy_on_urlopen()

        review_request.publish(review_request.submitter)

        self.assertTrue(urlopen.spy.called)
        self.assertEqual(data['url'],
                         'https://circleci.com/api/v1.1/project/github/'
                         'mypublicorg/mypublicorgrepo/tree/review-requests?'
                         'circle-token=test-api-token')

        self.assertEqual(data['build_params']['CIRCLE_JOB'], 'reviewboard')
        self.assertIn('REVIEWBOARD_API_TOKEN', data['build_params'])
        self.assertIn('REVIEWBOARD_DIFF_REVISION', data['build_params'])
        self.assertIn('REVIEWBOARD_REVIEW_REQUEST', data['build_params'])
        self.assertIn('REVIEWBOARD_STATUS_UPDATE_ID', data['build_params'])

        self.assertEqual(data['build_params']['REVIEWBOARD_SERVER'],
                         'http://example.com/s/%s/' % self.local_site_name)
        self.assertEqual(data['build_params']['REVIEWBOARD_LOCAL_SITE'],
                         self.local_site_name)

    def test_manual_run_no_build_on_publish(self):
        """Testing lack of CircleCIIntegration build when a new review
        request is made with the run manually configuration
        """
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config(run_manually=True)
        self.integration.enable_integration()

        data = {}

        def _urlopen(request, **kwargs):
            # We can't actually do any assertions in here, because they'll get
            # swallowed by SignalHook's sandboxing. We therefore record the
            # data we need and assert later.
            data['url'] = request.get_full_url()
            data['build_params'] = json.loads(request.data)['build_parameters']

            class _Response(object):
                def read(self):
                    return json.dumps({
                        'build_url': 'http://example.com/gh/org/project/35',
                    })

            return _Response()

        self.spy_on(urlopen, call_fake=_urlopen)

        review_request.publish(review_request.submitter)

        self.assertFalse(urlopen.spy.called)
        self.assertEqual(data, {})

    def test_build_manual_run(self):
        """Testing CircleCIIntegration build via a manual trigger"""
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config(run_manually=True)
        self.integration.enable_integration()

        status_update = \
            self.create_status_update(service_id='circle-ci',
                                      review_request=review_request)

        data = {}

        def _urlopen(request, **kwargs):
            # We can't actually do any assertions in here, because they'll get
            # swallowed by SignalHook's sandboxing. We therefore record the
            # data we need and assert later.
            data['url'] = request.get_full_url()
            data['build_params'] = json.loads(request.data)['build_parameters']

            class _Response(object):
                def read(self):
                    return json.dumps({
                        'build_url': 'http://example.com/gh/org/project/35',
                    })

            return _Response()

        self.spy_on(urlopen, call_fake=_urlopen)

        status_update_request_run.send(sender=self.__class__,
                                       status_update=status_update)

        self.assertTrue(urlopen.spy.called)
        self.assertEqual(data['url'],
                         'https://circleci.com/api/v1.1/project/github/'
                         'mypublicorg/mypublicorgrepo/tree/review-requests?'
                         'circle-token=test-api-token')

        self.assertEqual(data['build_params']['CIRCLE_JOB'], 'reviewboard')
        self.assertIn('REVIEWBOARD_API_TOKEN', data['build_params'])
        self.assertIn('REVIEWBOARD_DIFF_REVISION', data['build_params'])
        self.assertIn('REVIEWBOARD_REVIEW_REQUEST', data['build_params'])
        self.assertIn('REVIEWBOARD_SERVER', data['build_params'])
        self.assertIn('REVIEWBOARD_STATUS_UPDATE_ID', data['build_params'])

    def test_non_github_review_request(self):
        """Testing CircleCIIntegration skipping a non-GitHub review request"""
        repository = self._create_repository(github=False)
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(urlopen, call_original=False)

        review_request.publish(review_request.submitter)

        self.assertFalse(urlopen.spy.called)

    def test_build_new_review_request_with_public_github_repository(self):
        """Testing CircleCIIntegration builds for a new review request with
        a public GitHub repository"""
        repository = self._create_repository(github=True,
                                             repository_plan='public')
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_urlopen()

        review_request.publish(review_request.submitter)

        self.assertTrue(urlopen.spy.called)
        self.assertEqual(data['url'],
                         'https://circleci.com/api/v1.1/project/github/'
                         'myuser/mypublicrepo/tree/review-requests?'
                         'circle-token=test-api-token')

    def test_build_new_review_request_with_private_github_repository(self):
        """Testing CircleCIIntegration builds for a new review request with
        a private GitHub repository"""
        repository = self._create_repository(github=True,
                                             repository_plan='private')
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_urlopen()

        review_request.publish(review_request.submitter)

        self.assertTrue(urlopen.spy.called)
        self.assertEqual(data['url'],
                         'https://circleci.com/api/v1.1/project/github/'
                         'myuser/myprivaterepo/tree/review-requests?'
                         'circle-token=test-api-token')

    def test_build_new_review_request_with_private_org_github_repository(self):
        """Testing CircleCIIntegration builds for a new review request with
        a private organization GitHub repository"""
        repository = self._create_repository(github=True,
                                             repository_plan='private-org')
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_urlopen()

        review_request.publish(review_request.submitter)

        self.assertTrue(urlopen.spy.called)
        self.assertEqual(data['url'],
                         'https://circleci.com/api/v1.1/project/github/'
                         'myprivateorg/myprivateorgrepo/tree/review-requests?'
                         'circle-token=test-api-token')

    def _create_repository(self,
                           github=True,
                           with_local_site=False,
                           repository_plan='public-org'):
        """Create and return a repository for testing.

        Args:
            github (bool, optional):
                Whether the repository should use the GitHub hosting service.

            with_local_site (bool, optional):
                Whether to limit the config to a local site.

            repository_plan (unicode, optional):
                The type of GitHub repository plan.

        Returns:
            reviewboard.scmtools.models.Repository:
            A repository for use in unit tests.
        """
        if github:
            account = HostingServiceAccount(service_name='github',
                                            username='myuser')

            def _http_get_user(_self, url, *args, **kwargs):
                self.assertEqual(url, 'https://api.github.com/user')

                payload = b'{}'
                headers = {
                    str('X-OAuth-Scopes'): str('admin:repo_hook, repo, user'),
                }

                return HostingServiceHTTPResponse(
                    request=HostingServiceHTTPRequest(url=url),
                    url=url,
                    data=payload,
                    headers=headers,
                    status_code=200)

            service = account.service
            self.spy_on(service.client.http_get,
                        call_fake=_http_get_user)

            service.authorize('myuser', 'mypass', None)
            self.assertTrue(account.is_authorized)

            service.client.http_get.unspy()

            repository = self.create_repository(
                with_local_site=with_local_site)
            repository.hosting_account = account
            repository.extra_data['repository_plan'] = repository_plan

            if repository_plan == 'public':
                repository.extra_data['github_public_repo_name'] = \
                    'mypublicrepo'
            elif repository_plan == 'public-org':
                repository.extra_data['github_public_org_name'] = 'mypublicorg'
                repository.extra_data['github_public_org_repo_name'] = \
                    'mypublicorgrepo'
            elif repository_plan == 'private':
                repository.extra_data['github_private_repo_name'] = \
                    'myprivaterepo'
            elif repository_plan == 'private-org':
                repository.extra_data['github_private_org_name'] = \
                    'myprivateorg'
                repository.extra_data['github_private_org_repo_name'] = \
                    'myprivateorgrepo'

            repository.save()
            return repository
        else:
            return self.create_repository()

    def _create_config(self, with_local_site=False, run_manually=False):
        """Create an integration config.

        Args:
            with_local_site (bool, optional):
                Whether to limit the config to a local site.

            run_manually (bool, optional):
                Whether to run CircleCIIntegration manually.
        """
        choice = ReviewRequestRepositoriesChoice()

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator('any'))
        ])

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        config = self.integration.create_config(name='Config 1',
                                                enabled=True,
                                                local_site=local_site)
        config.set('conditions', condition_set.serialize())
        config.set('branch_name', 'review-requests')
        config.set('run_manually', run_manually)
        config.set('circle_api_token', 'test-api-token')
        config.save()

        return config

    def _spy_on_urlopen(self):
        """Wrapper function for spying on the urlopen function.

        Returns:
            dict:
            Faked response from CircleCI.
        """
        data = {}

        def _urlopen(request, **kwargs):
            # We can't actually do any assertions in here, because they'll get
            # swallowed by SignalHook's sandboxing. We therefore record the
            # data we need and assert later.
            data['url'] = request.get_full_url()
            data['build_params'] = json.loads(request.data)['build_parameters']

            class _Response(object):
                def read(self):
                    return json.dumps({
                        'build_url': 'http://example.com/gh/org/project/35',
                    }).encode('utf-8')

            return _Response()

        self.spy_on(urlopen, call_fake=_urlopen)

        return data



"""Integration for building changes on Travis CI."""

from __future__ import annotations

import base64
import logging
from typing import List, TYPE_CHECKING

import yaml
from django.urls import reverse
from django.utils.functional import cached_property
from reviewboard.admin.server import build_server_url

from rbintegrations.baseci.integration import (BaseCIIntegration,
                                               BuildPrepData)
from rbintegrations.travisci.api import TravisAPI
from rbintegrations.travisci.forms import TravisCIIntegrationConfigForm
from rbintegrations.util.compat.logs import log_timed

if TYPE_CHECKING:
    from reviewboard.integrations.models import IntegrationConfig
    from reviewboard.reviews.models import StatusUpdate


logger = logging.getLogger(__name__)


class TravisCIIntegration(BaseCIIntegration):
    """Integrates Review Board with Travis CI."""

    name = 'Travis CI'
    description = 'Builds diffs posted to Review Board using Travis CI.'
    config_form_cls = TravisCIIntegrationConfigForm

    status_update_service_id = 'travis-ci'

    bot_username = 'travis-ci'
    bot_user_first_name = 'Travis'
    bot_user_last_name = 'CI'

    @cached_property
    def icon_static_urls(self):
        """Return the icons used for the integration.

        Returns:
            dict:
            The icons for Travis CI.
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        return {
            '1x': extension.get_static_url('images/travisci/icon.png'),
            '2x': extension.get_static_url('images/travisci/icon@2x.png'),
        }

    def prepare_builds(
        self,
        prep_data: BuildPrepData,
    ) -> bool:
        """Prepare for builds.

        This will check if the change is on a supported hosting service
        (GitHub), and only permit the build if it is.

        Version Added:
            3.1

        Args:
            prep_data (BuildPrepData):
                The builds preparation data to modify.

        Returns:
            bool:
            ``True`` if the builds should proceed. ``False`` if builds should
            be skipped.
        """
        # Only build changes against GitHub repositories.
        repository = prep_data.repository

        if not (repository and
                repository.hosting_account and
                repository.hosting_account.service_name == 'github'):
            return False

        scmtool = repository.get_scmtool()
        review_request = prep_data.review_request

        prep_data.extra_state.update({
            'commit_message': '%s\n\n%s' % (review_request.summary,
                                            review_request.description),
            'diff_data': (
                base64.b64encode(
                    scmtool.get_parser(b'').raw_diff(prep_data.diffset))
                .decode('utf-8')
            ),
            'webhook_url': build_server_url(reverse('travis-ci-webhook')),
        })

        return True

    def start_build(
        self,
        *,
        prep_data: BuildPrepData,
        config: IntegrationConfig,
        status_update: StatusUpdate,
    ) -> None:
        """Start a new build.

        This will trigger a Travis-CI build for the given review request and
        configuration.

        Version Added:
            3.1

        Args:
            prep_data (BuildPrepData):
                The builds preparation data containing information for the
                build.

            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration for the integration triggering this build.

            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update indicating the status of this build.

        Raises:
            rbintegrations.baseci.errors.CIBuildError:
                There was an error invoking the build.
        """
        diffset = prep_data.diffset

        travis_config = yaml.load(config.get('travis_yml'),
                                  Loader=yaml.SafeLoader)

        # Add set-up and patching to the start of the "before_install"
        # section of the config.
        before_install: List[str] = []

        if travis_config.get('git', {}).get('depth', True) is not False:
            before_install.append('git fetch --unshallow origin || true')

        before_install.append('git checkout %s' % diffset.base_commit_id)

        # Add parent diff if necessary
        parent_diff_data = self._get_parent_diff(diffset)

        if parent_diff_data:
            parent_diff_data = base64.b64encode(parent_diff_data)
            before_install.append('echo %s | base64 --decode | patch -p1' %
                                  parent_diff_data.decode('utf-8'))

        before_install.append('echo %s | base64 --decode | patch -p1' %
                              prep_data.extra_state['diff_data'])

        old_install = travis_config.get('before_install', [])

        if not isinstance(old_install, list):
            old_install = [old_install]

        travis_config['before_install'] = before_install + old_install

        # Set up webhook notifications.
        notifications = travis_config.get('notifications') or {}
        webhooks = notifications.get('webhooks') or {}

        urls = webhooks.get('urls', [])

        if not isinstance(urls, list):
            urls = [urls]

        urls.append(prep_data.extra_state['webhook_url'])

        webhooks['urls'] = urls
        webhooks['on_start'] = 'always'

        notifications['webhooks'] = webhooks
        notifications['email'] = False
        travis_config['notifications'] = notifications

        # Add some special data in the environment so that when the
        # webhooks come in, we can find the right status update to update.
        env = travis_config.setdefault('env', {})

        if not isinstance(env, dict):
            env = {
                'matrix': env,
            }

        global_ = env.setdefault('global', [])

        if not isinstance(global_, list):
            global_ = [global_]

        global_ += [
            'REVIEWBOARD_STATUS_UPDATE_ID=%d' % status_update.pk,
            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d' % config.pk,
        ]

        env['global'] = global_

        travis_config['env'] = env

        # Time to kick off the build!
        with log_timed(f'Triggering Travis CI build for review request '
                       f'{prep_data.review_request.get_absolute_url()} '
                       f'(diffset revision {diffset.revision})',
                       logger=logger):
            repo_slug = self._get_github_repository_slug(prep_data.repository)

            api = TravisAPI(config)
            api.start_build(
                repo_slug=repo_slug,
                travis_config=travis_config,
                commit_message=prep_data.extra_state['commit_message'],
                branch=config.get('branch_name') or 'master')

    def _get_github_repository_slug(self, repository):
        """Return the "slug" for a GitHub repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository.

        Returns:
            unicode:
            The slug for use with the Travis CI API.
        """
        extra_data = repository.extra_data
        plan = extra_data['repository_plan']

        if plan == 'public':
            return '%s/%s' % (repository.hosting_account.username,
                              extra_data['github_public_repo_name'])
        elif plan == 'public-org':
            return '%s/%s' % (extra_data['github_public_org_name'],
                              extra_data['github_public_org_repo_name'])
        elif plan == 'private':
            return '%s/%s' % (repository.hosting_account.username,
                              extra_data['github_private_repo_name'])
        elif plan == 'private-org':
            return '%s/%s' % (extra_data['github_private_org_name'],
                              extra_data['github_private_org_repo_name'])
        else:
            raise ValueError('Unexpected plan for GitHub repository %d: %s'
                             % (repository.pk, plan))

    def _get_parent_diff(self, diffset):
        """Return the raw parent diff.

        Args:
            diffset (reviewboard.diffviewer.models.DiffSet):
                The diffset to get the parent data for.

        Returns:
            bytes:
            The raw parent diff data, if available. None if not.
        """
        data = []

        for filediff in diffset.files.all():
            parent_diff = filediff.parent_diff

            if parent_diff:
                data.append(parent_diff)

        if data:
            return b''.join(data)
        else:
            return None



"""The form for configuring the GitLab CI integration.

Version Added:
    5.0
"""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

import gitlab
from django import forms
from django.core.exceptions import ValidationError
from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _

from rbintegrations.baseci.forms import BaseCIIntegrationConfigForm

if TYPE_CHECKING:
    from typing import Any

    from rbintegrations.gitlabci.integration import GitLabCIIntegration


logger = logging.getLogger(__name__)


def validate_json(
    value: str,
) -> None:
    """Validate that a given string is parseable as JSON.

    Version Added:
        5.0

    Args:
        value (str):
            The value to validate.

    Raises:
        django.core.exceptions.ValidationError:
            The value could not be parsed.
    """
    try:
        json.loads(value)
    except Exception as e:
        raise ValidationError(
            _('Unable to parse JSON: %s') % str(e))


class TokenChoices(TextChoices):
    """The choices for the gitlab_token_type field.

    Version Added:
        5.0
    """

    PRIVATE_TOKEN = 'private_token', _('Private API token')
    TRIGGER_TOKEN = 'trigger_token', _('Trigger token')


class GitLabCIIntegrationConfigForm(BaseCIIntegrationConfigForm):
    """Form for configuring the GitLab CI integration.

    Version Added:
        5.0
    """

    gitlab_endpoint = forms.URLField(
        label=_('Server'),
        help_text=_(
            'The URL to your GitLab server. For example, '
            '<code>https://gitlab.example.com/</code>'
        ),
        widget=forms.widgets.TextInput(attrs={
            'size': 40,
        }),
    )

    gitlab_token_type = forms.ChoiceField(
        label=_('Token type'),
        help_text=_('The type of access token used to create pipelines.'),
        choices=TokenChoices.choices,
    )

    gitlab_token = forms.CharField(
        label=_('Token'),
        help_text=_('The API token used for authentication.'),
        widget=forms.PasswordInput(render_value=True),
    )

    gitlab_name = forms.CharField(
        label=_('Project name or ID'),
        initial='{repository_name}',
        help_text=_(
            'Name or ID (required for trigger token). This can include the '
            'following variables: <code>{repository_name}</code>.'
        ),
    )

    gitlab_ref = forms.CharField(
        label=_('Git refname'),
        initial='{branch}',
        help_text=_(
            'Branch, tag, or other Git refname (if not equal to the review '
            'request "branch" field). This can include the following '
            'variables: <code>{branch}</code>.'
        ),
    )

    gitlab_inputs = forms.CharField(
        label=_('Pipeline inputs'),
        initial='{}',
        help_text=_(
            'Inputs for the GitLab CI pipeline. This should be formatted as '
            'a JSON object. Values in this can use the special variables '
            '<code>{branch}</code> and <code>{repository_name}</code>.'
        ),
        widget=forms.widgets.Textarea(attrs={
            'cols': '80',
            'rows': '10',
        }),
        validators=[validate_json],
    )

    gitlab_vars = forms.CharField(
        label=_('Additional variables'),
        initial='{}',
        help_text=_(
            'Additional variables to provide to the GitLab CI pipeline. This '
            'should be formatted as a JSON object. Values in this can use the '
            'special variables <code>{branch}</code> and '
            '<code>{repository_name}</code>.'
        ),
        widget=forms.widgets.Textarea(attrs={
            'cols': '80',
            'rows': '10',
        }),
        validators=[validate_json],
    )

    gitlab_report_job_state = forms.BooleanField(
        label=_('Report job state'),
        required=False,
        help_text=_(
            'If checked, GitLab results will include a summary of all jobs '
            'within the pipeline.'
        ),
    )

    gitlab_webhook_secret_token = forms.CharField(
        label=_('GitLab WebHook secret token'),
        required=False,
        help_text=_(
            'The secret token configured in the GitLab Webhook '
            'configuration. This is optional, but setting this both here and '
            'in the Webhook configuration can ensure '
        ),
        widget=forms.PasswordInput(render_value=True),
    )

    ######################
    # Instance variables #
    ######################

    #: The integration being configured.
    integration: GitLabCIIntegration

    def clean(self) -> dict[str, Any] | None:
        """Clean the form.

        This validates the user credentials that have been provided.

        Returns:
            dict:
            The cleaned form data.
        """
        cleaned_data = super().clean()

        if not self.is_valid() or not cleaned_data:
            # Form validation has already failed.
            return cleaned_data

        assert cleaned_data is not None

        try:
            if cleaned_data['gitlab_token_type'] == 'private_token':
                gl = gitlab.Gitlab(
                    url=cleaned_data['gitlab_endpoint'],
                    private_token=cleaned_data['gitlab_token'],
                )
                gl.auth()
        except gitlab.GitlabError as e:
            self.add_error('gitlab_endpoint', str(e))

        return cleaned_data

    class Meta(BaseCIIntegrationConfigForm.Meta):
        """Metadata for the form."""

        fieldsets = (
            (_('What To Build'), {
                'description': _(
                    'You can choose which review requests to build using this '
                    'configuration by setting conditions here. At a minimum, '
                    'this should include the specific repository to use '
                    'this configuration for.'
                ),
                'fields': (
                    'conditions',
                ),
            }),
            (_('Where To Build'), {
                'description': _(
                    'Configure the address and authentication credentials '
                    'for the GitLab server handling your builds.'
                ),
                'fields': (
                    'gitlab_endpoint',
                    'gitlab_token_type',
                    'gitlab_token',
                ),
            }),
            (_('How To Build'), {
                'fields': (
                    'gitlab_name',
                    'gitlab_ref',
                    'gitlab_inputs',
                    'gitlab_vars',
                ),
            }),
            (_('When To Build'), {
                'fields': (
                    'run_manually',
                    'timeout',
                ),
            }),
            (_('Reporting Results'), {
                'fields': (
                    'gitlab_report_job_state',
                    'gitlab_webhook_secret_token',
                ),
            }),
        )



"""Utility functions for I Done This integration."""

import json
import logging
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from django.core.cache import cache
from djblets.cache.backend import cache_memoize, make_cache_key
from reviewboard.scmtools.crypto_utils import decrypt_password


IDONETHIS_API_BASE_URL = 'https://beta.idonethis.com/api/v2'
TEAM_IDS_CACHE_EXPIRATION = 24 * 60 * 60  # 1 day


def create_idonethis_request(request_path, api_token, json_payload=None):
    """Create a urllib request for the I Done This API.

    Args:
        request_path (unicode):
            The API request path, relative to the base API URL.

        api_token (unicode):
            The user's API token for authorization.

        json_payload (unicode, optional):
            JSON payload for a POST request. If this is omitted,
            the request will be a GET.

    Returns:
        urllib2.Request:
        The I Done This API request with the provided details.
    """
    url = '%s/%s' % (IDONETHIS_API_BASE_URL, request_path)
    headers = {
        'Authorization': 'Token %s' % api_token,
    }

    if json_payload is not None:
        headers['Content-Type'] = 'application/json'
        json_payload = json_payload.encode('utf-8')

    return Request(url, json_payload, headers)


def get_user_api_token(user):
    """Return the user's API token for I Done This.

    Args:
        user (django.contrib.auth.models.User):
            The user whose API token should be retrieved.

    Returns:
        unicode:
        The user's API token, or ``None`` if the user has not set one.
    """
    try:
        settings = user.get_profile().settings['idonethis']
        return decrypt_password(settings['api_token'])
    except KeyError:
        return None


def get_user_team_ids(user):
    """Return a set of I Done This team IDs that the user belongs to.

    Retrieves the set of teams from the I Done This API and caches it to
    avoid excessive requests. Team membership is not expected to change
    frequently, but the cache can be manually deleted if necessary.

    Args:
        user (django.contrib.auth.models.User):
            The user whose cached team IDs should be retrieved.

    Returns:
        set:
        The user's team IDs, or ``None`` if they could not be retrieved.
    """
    def _get_user_team_ids_uncached():
        request = create_idonethis_request(request_path='teams',
                                           api_token=api_token)
        logging.debug('IDoneThis: Loading teams for user "%s", '
                      'request "%s %s"',
                      user.username,
                      request.get_method(),
                      request.get_full_url())

        try:
            teams_data = urlopen(request).read()
        except (HTTPError, URLError) as e:
            if isinstance(e, HTTPError):
                error_info = '%s, error data: %s' % (e, e.read())
            else:
                error_info = e.reason

            logging.error('IDoneThis: Failed to load teams for user "%s", '
                          'request "%s %s": %s',
                          user.username,
                          request.get_method(),
                          request.get_full_url(),
                          error_info)
            raise

        return set(t['hash_id'] for t in json.loads(teams_data))

    api_token = get_user_api_token(user)

    if not api_token:
        return None

    try:
        return set(cache_memoize(_make_user_team_ids_cache_key(user),
                                 _get_user_team_ids_uncached,
                                 expiration=TEAM_IDS_CACHE_EXPIRATION))
    except Exception as e:
        logging.error('IDoneThis: Failed to load teams for user "%s": %s',
                      user.username,
                      e)
        return None


def delete_cached_user_team_ids(user):
    """Delete the user's cached I Done This team IDs.

    Args:
        user (django.contrib.auth.models.User):
            The user whose cached team IDs should be deleted.
    """
    logging.debug('IDoneThis: Deleting cached team IDs for user "%s"',
                  user.username)
    cache.delete(make_cache_key(_make_user_team_ids_cache_key(user)))


def _make_user_team_ids_cache_key(user):
    """Make a cache key for the user's I Done This team IDs.

    Args:
        user (django.contrib.auth.models.User):
            The user to generate the cache key for.

    Returns:
        unicode:
        The cache key for the user's team IDs.
    """
    return 'idonethis-team_ids-%s' % user.username



"""Unit tests for the Microsoft Teams integration.

Version Added:
    4.0
"""

from __future__ import annotations

import json
from typing import List, Optional, TYPE_CHECKING
from urllib.request import urlopen

import kgb
from django.contrib.auth.models import User
from djblets.conditions import ConditionSet, Condition
from djblets.testing.decorators import add_fixtures
from reviewboard.accounts.trophies import TrophyType, trophies_registry
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.models import ReviewRequestDraft

from rbintegrations.msteams.integration import MSTeamsIntegration
from rbintegrations.testing.testcases import IntegrationTestCase

if TYPE_CHECKING:
    from djblets.integrations.models import BaseIntegrationConfig
    from djblets.util.typing import JSONDict


class MSTeamsIntegrationTests(IntegrationTestCase):
    """Tests Review Board integration with Microsoft Teams.

    Version Added:
        4.0
    """

    integration_cls = MSTeamsIntegration
    fixtures = ['test_scmtools', 'test_users']

    def setUp(self) -> None:
        super(MSTeamsIntegrationTests, self).setUp()

        self.user = User.objects.create(username='test',
                                        first_name='Test',
                                        last_name='User')

        self.spy_on(urlopen, call_original=False)
        self.spy_on(self.integration.notify)

    def test_notify_new_review_request(self) -> None:
        """Testing MSIntegration notifies on new review request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    @add_fixtures(['test_site'])
    def test_notify_new_review_request_with_local_site(self) -> None:
        """Testing MSTeamsIntegration notifies on new review request
        with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            target_people=[self.user],
            publish=False)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from [Test User]'
                     '(http://example.com/s/local-site-1/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/s/local-site-1/r/1/)',
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_new_review_request_with_diff(self) -> None:
        """Testing MSTeamsIntegration notifies on new review request
        with diff
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)
        self.create_diffset(review_request)

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Diff',
                    'value': '[Revision 1](http://example.com/r/1/diff/1/)',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    @add_fixtures(['test_site'])
    def test_notify_new_review_request_with_local_site_and_diff(self) -> None:
        """Testing MSTeamsIntegration notifies on new review request
        with local site and diff
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=False)
        self.create_diffset(review_request)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from [Test User]'
                     '(http://example.com/s/local-site-1/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/s/local-site-1/r/1/)',
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Diff',
                    'value': '[Revision 1]'
                             '(http://example.com/s/local-site-1/r/1/diff/1/)',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_new_review_request_with_image_file_attachment(
        self,
    ) -> None:
        """Testing MSTeamsIntegration notifies on new review request
        with image file attachment
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)
        attachment = self.create_file_attachment(review_request)

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
            image_url=attachment.get_absolute_url(),
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_new_review_request_with_invalid_file_attachment(
        self,
    ) -> None:
        """Testing MSTeamsIntegration doesn't include an image url
        for a review request that has a file attachment with an invalid
        file extension
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)
        self.create_file_attachment(
            review_request,
            mimetype='text/plain',
            orig_filename='foo.txt')

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_new_review_request_with_fish_trophy(self) -> None:
        """Testing MSTeamsIntegration notifies on new review request
        with fish trophy
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False,
            id=12321)

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test/)',
            title='[#12321: Test Review Request]'
                  '(http://example.com/r/12321/)',
            thumb_url=self.integration.trophy_urls['fish'],
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_new_review_request_with_milestone_trophy(self) -> None:
        """Testing MSTeamsIntegration notifies on new review request
        with milestone trophy
        """
        review_request = self.create_review_request(
            create_repository=True,
            id=10000,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test/)',
            title='[#10000: Test Review Request]'
                  '(http://example.com/r/10000/)',
            thumb_url=self.integration.trophy_urls['milestone'],
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_new_review_request_with_custom_trophy(self) -> None:
        """Testing MSTeamsIntegration notifies on new review request
        with ignored custom trophy
        """
        class MyTrophy(TrophyType):
            category = 'test'

            def __init__(self) -> None:
                super(MyTrophy, self).__init__(title='My Trophy',
                                               image_url='blahblah')

            def qualifies(self, review_request):
                return True

        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        trophies_registry.register(MyTrophy)

        try:
            review_request.publish(self.user)
        finally:
            trophies_registry.unregister(MyTrophy)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_updated_review_request(self) -> None:
        """Testing MSTeamsIntegration notifies on updated review
        request"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=True)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New update from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: My new summary](http://example.com/r/1/)',
            fields=[
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_updated_review_request_with_change_description(
        self,
    ) -> None:
        """Testing MSTeamsIntegration notifies on updated review request
        with change description
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=True)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        changedesc = draft.changedesc
        changedesc.text = 'These are my changes.'
        changedesc.save()

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New update from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: My new summary](http://example.com/r/1/)',
            body='These are my changes.',
            fields=[
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_updated_review_request_with_new_image_attachments(
        self,
    ) -> None:
        """Testing MSTeamsIntegration notifies on updated review
        request with new image attachments
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            target_people=[self.user],
            publish=False)
        self.create_file_attachment(review_request)
        review_request.publish(self.user)

        attachment = self.create_file_attachment(review_request,
                                                 caption='My new attachment',
                                                 draft=True)

        self._create_config()
        self.integration.enable_integration()

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New update from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
            image_url=attachment.get_absolute_url(),
            fields=[
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def test_notify_closed_review_request_as_submitted(self) -> None:
        """Testing MSTeamsIntegration notifies on closing review
        request as submitted
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)

        self._create_config()
        self.integration.enable_integration()

        review_request.close(review_request.SUBMITTED)

        self._check_notify_request(
            pre_text='Closed as completed by '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
        )

    def test_notify_closed_review_request_as_discarded(self) -> None:
        """Testing MSTeamsIntegration notifies on closing review
        request as discarded
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)

        self._create_config()
        self.integration.enable_integration()

        review_request.close(review_request.DISCARDED)

        self._check_notify_request(
            pre_text='Discarded by '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
        )

    @add_fixtures(['test_site'])
    def test_notify_closed_review_request_with_local_site(self) -> None:
        """Testing MSTeamsIntegration notifies on closing review
        request with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        review_request.close(review_request.SUBMITTED)

        self._check_notify_request(
            pre_text='Closed as completed by [Test User]'
                     '(http://example.com/s/local-site-1/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/s/local-site-1/r/1/)',
        )

    def test_notify_reopened_review_request(self) -> None:
        """Testing MSTeamsIntegration notifies on reopened review
        request
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=True)
        review_request.close(review_request.SUBMITTED)

        self._create_config()
        self.integration.enable_integration()

        review_request.reopen(self.user)

        self._check_notify_request(
            pre_text='Reopened by '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request](http://example.com/r/1/)',
            body='My description.',
        )

    @add_fixtures(['test_site'])
    def test_notify_reopened_review_request_with_local_site(self) -> None:
        """Testing MSTeamsIntegration notifies on reopened review
        request with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            with_local_site=True,
            local_id=1,
            publish=True)
        review_request.close(review_request.SUBMITTED)

        review_request.local_site.users.add(self.user)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        review_request.reopen(self.user)

        self._check_notify_request(
            pre_text='Reopened by [Test User]'
                     '(http://example.com/s/local-site-1/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/s/local-site-1/r/1/)',
            body='My description.',
        )

    def test_notify_new_review_with_body_top(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with
        body_top
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    body_top='This is my review.')
        self.create_general_comment(review)

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='This is my review.'
        )

    @add_fixtures(['test_site'])
    def test_notify_new_review_with_local_site(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with local
        site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        review = self.create_review(review_request,
                                    user=self.user,
                                    body_top='This is my review.')
        self.create_general_comment(review)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from [Test User]'
                     '(http://example.com/s/local-site-1/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/s/local-site-1/r/1/#review1)',
            body='This is my review.'
        )

    def test_notify_new_review_with_comments(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with only
        comments
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.')

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='My general comment.'
        )

    def test_notify_new_review_with_one_open_issue(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with 1 open
        issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment.',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='My general comment.',
            fields=[
                {
                    'title': 'Open Issues',
                    'value': '⚠ 1 issue',
                },
            ],
        )

    def test_notify_new_review_with_open_issues(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with > 1
        open issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user,
                                    body_top='')
        self.create_general_comment(review, text='My general comment 1.',
                                    issue_opened=True)
        self.create_general_comment(review, text='My general comment 2.',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='My general comment 1.',
            fields=[
                {
                    'title': 'Open Issues',
                    'value': '⚠ 2 issues',
                },
            ],
        )

    def test_notify_new_review_with_ship_it(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with
        Ship It!
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='Test Body Bottom',
            fields=[
                {
                    'title': 'Ship it!',
                    'value': '✅',
                },
            ],
        )

    def test_notify_new_review_with_ship_it_and_custom_body_top(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with Ship
        It and custom body_top
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='This is body_top.')
        self.create_general_comment(review)

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='This is body_top.',
            fields=[
                {
                    'title': 'Ship it!',
                    'value': '✅',
                },
            ],
        )

    def test_notify_new_review_with_ship_it_and_one_open_issue(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with Ship
        It! and 1 open issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review, issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='My comment',
            fields=[
                {
                    'title': 'Fix it, then Ship it!',
                    'value': '⚠ 1 issue',
                },
            ],
        )

    def test_notify_new_review_with_ship_it_and_open_issues(self) -> None:
        """Testing MSTeamsIntegration notifies on new review with Ship
        It! and > 1 open issues
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True,
                                    body_top='Ship It!')
        self.create_general_comment(review, text='My general comment 1.',
                                    issue_opened=True)
        self.create_general_comment(review, text='My general comment 2.',
                                    issue_opened=True)

        self._create_config()
        self.integration.enable_integration()

        review.publish()

        self._check_notify_request(
            pre_text='New review from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='My general comment 1.',
            fields=[
                {
                    'title': 'Fix it, then Ship it!',
                    'value': '⚠ 2 issues',
                },
            ],
        )

    def test_notify_new_reply_with_body_top(self) -> None:
        """Testing MSTeamsIntegration notifies on new reply with
        body_top
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review,
                                  user=self.user,
                                  body_top='This is body_top.')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config()
        self.integration.enable_integration()

        reply.publish()

        self._check_notify_request(
            pre_text='New reply from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#review1)',
            body='This is body_top.',
        )

    @add_fixtures(['test_site'])
    def test_notify_new_reply_with_local_site(self) -> None:
        """Testing MSTeamsIntegration notifies on new reply with local
        site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)

        review_request.local_site.users.add(self.user)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review,
                                  user=self.user,
                                  body_top='This is body_top.')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config(with_local_site=True)
        self.integration.enable_integration()

        reply.publish()

        self._check_notify_request(
            pre_text='New reply from [Test User]'
                     '(http://example.com/s/local-site-1/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/s/local-site-1/r/1/#review1)',
            body='This is body_top.',
        )

    def test_notify_new_reply_with_comment(self) -> None:
        """Testing MSTeamsIntegration notifies on new reply with
        comment
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review, user=self.user, body_top='')
        self.create_general_comment(reply,
                                    text='This is a comment.',
                                    reply_to=comment)

        self._create_config()
        self.integration.enable_integration()

        reply.publish()

        self._check_notify_request(
            pre_text='New reply from '
                     '[Test User](http://example.com/users/test/)',
            title='[#1: Test Review Request]'
                  '(http://example.com/r/1/#gcomment2)',
            body='This is a comment.',
        )

    def test_notify_with_no_webhook_url(self) -> None:
        """Testing MSTeamsIntegration logs an error when attempting to send
        a notification when no webhook URL is configured
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            description='My description.',
            publish=False)

        config = self._create_config()
        config.set('webhook_url', None)
        config.save()
        self.integration.enable_integration()

        with self.assertLogs() as logs:
            review_request.publish(self.user)

        self.assertRegex(
            logs.records[0].getMessage(),
            r'\[[a-z0-9-]+\] Failed to send notification: '
            r'WebHook URL has not been configured.')

    def test_notify_new_review_request_escaped_markdown(self) -> None:
        """Testing MSIntegration escapes []() characters when needed
        in its notification
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            # This would break a Markdown link.
            summary='Test](Foo)',
            description='My description.',
            target_people=[self.user],
            publish=False)

        self._create_config()
        self.integration.enable_integration()

        # This would break a Markdown link.
        self.spy_on(
            self.integration.get_user_text_url,
            op=kgb.SpyOpReturn('http://example.com/users/test?val=)&foo=(1)'))

        review_request.publish(self.user)

        self._check_notify_request(
            pre_text='New review request from '
                     '[Test User](http://example.com/users/test?'
                     'val=%29&foo=%281%29)',
            title='[#1: Test\\](Foo)](http://example.com/r/1/)',
            fields=[
                {
                    'title': 'Description',
                    'value': 'My description.',
                },
                {
                    'title': 'Repository',
                    'value': 'Test Repo',
                },
                {
                    'title': 'Branch',
                    'value': 'my-branch',
                },
            ],
        )

    def _create_config(
        self,
        with_local_site: bool = False
    ) -> BaseIntegrationConfig:
        """Set configuration values for MSTeamsIntegration.

        Args:
            with_local_site (bool, optional):
                Whether the configuration should be for a local site.

        Returns:
            djblets.integrations.models.BaseIntegrationConfig:
            The configuration for the integration.
        """
        choice = ReviewRequestRepositoriesChoice()

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator('any')),
        ])

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        config = self.integration.create_config(name='Config 1',
                                                enabled=True,
                                                local_site=local_site)
        config.set('notify_username', 'RB User')
        config.set('webhook_url', 'http://example.com/msteams-url/')
        config.set('conditions', condition_set.serialize())
        config.save()

        return config

    def _check_notify_request(
        self,
        *,
        pre_text,
        title,
        body: str = '',
        fields: Optional[List[JSONDict]] = None,
        thumb_url: Optional[str] = None,
        image_url: Optional[str] = None,
    ) -> None:
        """Check that a notify and HTTP request meets expected criteria.

        This will ensure that only a single request was invoked, and that the
        request information contains the appropriate headers, string types,
        and payload message content.

        Args:
            pre_text (str):
                The expected text to display before the rest of the message.

            title (str):
                The expected title of the message.

            body (str, optional):
                The expected body of the message.

            fields (list of djblets.util.typing.JSONDict):
                The expected fields to display in the message.

            thumb_url (str, optional):
                Expected URL of an image to show on the side of the message.

            image_url (str, optional):
                Expected URL of an image to show in the message.

        Raises:
            AssertionError:
                One or more of the checks failed.
        """
        pre_text_card = {
            'type': 'ColumnSet',
            'columns': [
                {
                    'type': 'Column',
                    'width': 'Auto',
                    'verticalContentAlignment': 'Center',
                    'items': [
                        {
                            'type': 'TextBlock',
                            'text': pre_text,
                            'size': 'Medium',
                            'style': 'Heading',
                        },
                    ]
                }
            ]
        }

        if thumb_url:
            pre_text_card['columns'].append({
                'type': 'Column',
                'width': 'auto',
                'spacing': 'None',
                'items': [
                    {
                        'type': 'Image',
                        'url': thumb_url,
                        'altText': 'A trophy earned on the review request.',
                        'width': '32px',
                        'height': '32px',
                    }
                ],
            })

        activity_card = {
            'type': 'ColumnSet',
            'columns': [
                {
                    'type': 'Column',
                    'width': 'Auto',
                    'items': [
                        {
                            'type': 'Image',
                            'url': self.integration.logo_url,
                            'alt': 'Review Board notification',
                            'width': '45px',
                            'height': '45px',
                        },
                    ],
                },
                {
                    'type': 'Column',
                    'items': [
                        {
                            'type': 'TextBlock',
                            'text': title,
                            'weight': 'Bolder',
                            'wrap': True,
                        },
                        {
                            'type': 'TextBlock',
                            'text': body,
                            'isSubtle': True,
                            'spacing': 'Small',
                            'wrap': True,
                        },
                    ]
                },
            ]
        }

        main_body = [
            pre_text_card,
            activity_card,
        ]

        if fields:
            main_body.append({
                'type': 'FactSet',
                'facts': [
                    {
                        'title': field['title'],
                        'value': field['value'],
                    }
                    for field in fields
                ],
            })

        if image_url:
            main_body.append({
                'type': 'Image',
                'url': image_url,
                'altText': 'An image file attachment from the review request.',
            })

        request = urlopen.last_call.args[0]
        request_body = request.data
        headers = request.headers

        self.assertSpyCallCount(self.integration.notify, 1)
        self.assertSpyCallCount(urlopen, 1)
        self.assertIsInstance(request_body, bytes)
        self.assertEqual(headers['Content-length'], str(len(request_body)))
        self.assertEqual(headers['Content-type'], str('application/json'))
        self.assertEqual(
            json.loads(request_body.decode('utf-8')),
            {
                'type': 'message',
                'attachments': [
                    {
                        'contentType':
                            'application/vnd.microsoft.card.adaptive',
                        'contentUrl': None,
                        'content':
                            {
                                '$schema': 'http://adaptivecards.io/schemas/'
                                           'adaptive-card.json',
                                'type': 'AdaptiveCard',
                                'version': '1.5',
                                'body': main_body,
                            },
                    },
                ],
            })



"""The form for configuring the Circle CI integration."""

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.reviews.conditions import ReviewRequestConditionChoices

from rbintegrations.baseci.forms import BaseCIIntegrationConfigForm


class CircleCIIntegrationConfigForm(BaseCIIntegrationConfigForm):
    """Form for configuring Circle CI."""

    circle_api_token = forms.CharField(
        label=_('API Token'),
        help_text=_('Your CircleCI API token. You can create these tokens '
                    'in the CircleCI repository settings under "API '
                    'Permissions".'))

    branch_name = forms.CharField(
        label=_('Build Branch'),
        required=False,
        help_text=_('An optional branch name to use for review request '
                    'builds within the CircleCI user interface.'))

    class Meta(BaseCIIntegrationConfigForm.Meta):
        fieldsets = (
            (_('What To Build'), {
                'description': _(
                    'You can choose which review requests to build using this '
                    'configuration by setting conditions here. At a minimum, '
                    'this should include the specific repository to use '
                    'this configuration for.'
                ),
                'fields': (
                    'conditions',
                ),
            }),
            (_('How To Build'), {
                'description': _(
                    "Builds performed on the code in review requests will use "
                    "a completely separate configuration from commits which "
                    "are pushed to the source code repository. The "
                    "configuration listed here will be used instead of the "
                    "contents of the repository's "
                    "<code>.circleci/config.yml</code> file. Note that "
                    "this should not contain any secret environment "
                    "variables."
                    "\n"
                    "It's also recommended to create a special branch head "
                    "in the repository to use for these builds, so they "
                    "don't appear to be happening on your main development "
                    "branch. This branch can contain anything (or even be "
                    "empty), since the code will come from the review "
                    "request."
                ),
                'fields': (
                    'circle_api_token',
                    'branch_name',
                ),
            }),
            (_('When To Build'), {
                'fields': (
                    'run_manually',
                    'timeout',
                ),
            }),
        )



"""Unit tests for the Asana integration."""

from __future__ import annotations

from django.http import HttpResponse
from reviewboard.site.urlresolvers import local_site_reverse

from rbintegrations.asana.integration import AsanaIntegration
from rbintegrations.asana.views import (AsanaTaskSearchView,
                                        AsanaWorkspaceListView)
from rbintegrations.testing.testcases import IntegrationTestCase


class AsanaIntegrationTests(IntegrationTestCase):
    """Tests for Asana."""

    integration_cls = AsanaIntegration
    fixtures = ['test_users', 'test_site']

    def test_workspace_list(self) -> None:
        """Testing AsanaWorkspaceListView"""
        self.spy_on(AsanaWorkspaceListView.get,
                    owner=AsanaWorkspaceListView,
                    call_fake=lambda self, request: HttpResponse(
                        '{}', content_type='application/json'))

        rsp = self.client.get(local_site_reverse('asana-workspace-list'))

        self.assertEqual(rsp.status_code, 200)

    def test_task_search(self) -> None:
        """Testing AsanaTaskSearchView"""
        self.spy_on(AsanaTaskSearchView.get,
                    owner=AsanaTaskSearchView,
                    call_fake=lambda self, request, **kwargs: HttpResponse(
                        '{}', content_type='application/json'))

        review_request = self.create_review_request(public=True)
        rsp = self.client.get(local_site_reverse(
            'asana-task-search',
            kwargs={
                'review_request_id': review_request.display_id,
            }))

        self.assertEqual(rsp.status_code, 200)

    def test_task_search_unpublished(self) -> None:
        """Testing AsanaTaskSearchView with an unpublished review request"""
        self.spy_on(AsanaTaskSearchView.get,
                    owner=AsanaTaskSearchView,
                    call_fake=lambda self, request, **kwargs: HttpResponse(
                        '{}', content_type='application/json'))

        review_request = self.create_review_request(public=False)
        self.client.login(username='dopey', password='dopey')
        rsp = self.client.get(local_site_reverse(
            'asana-task-search',
            kwargs={
                'review_request_id': review_request.display_id,
            }))

        self.assertEqual(rsp.status_code, 403)

    def test_task_search_with_local_site(self) -> None:
        """Testing AsanaTaskSearchView with a Local Site"""
        self.spy_on(AsanaTaskSearchView.get,
                    owner=AsanaTaskSearchView,
                    call_fake=lambda self, request, **kwargs: HttpResponse(
                        '{}', content_type='application/json'))

        self.client.login(username='doc', password='doc')
        review_request = self.create_review_request(public=True,
                                                    with_local_site=True)
        rsp = self.client.get(local_site_reverse(
            'asana-task-search',
            local_site_name=review_request.local_site.name,
            kwargs={
                'review_request_id': review_request.display_id,
            }))

        self.assertEqual(rsp.status_code, 200)

    def test_task_search_with_local_site_no_access(self) -> None:
        """Testing AsanaTaskSearchView with a Local Site that the user does not
        have access to
        """
        self.spy_on(AsanaTaskSearchView.get,
                    owner=AsanaTaskSearchView,
                    call_fake=lambda self, request, **kwargs: HttpResponse(
                        '{}', content_type='application/json'))

        self.client.login(username='dopey', password='dopey')
        review_request = self.create_review_request(public=True,
                                                    with_local_site=True)
        rsp = self.client.get(local_site_reverse(
            'asana-task-search',
            local_site_name=review_request.local_site.name,
            kwargs={
                'review_request_id': review_request.display_id,
            }))

        self.assertEqual(rsp.status_code, 403)



"""Integration with I Done This."""

import json
import logging
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from reviewboard.admin.server import build_server_url
from reviewboard.extensions.hooks import AccountPagesHook, SignalHook
from reviewboard.integrations.base import Integration
from reviewboard.reviews.models import BaseComment, ReviewRequest
from reviewboard.reviews.signals import (review_request_closed,
                                         review_request_published,
                                         review_request_reopened,
                                         review_published,
                                         reply_published)

from rbintegrations.idonethis import entries
from rbintegrations.idonethis.forms import IDoneThisIntegrationConfigForm
from rbintegrations.idonethis.pages import IDoneThisIntegrationAccountPage
from rbintegrations.idonethis.utils import (create_idonethis_request,
                                            get_user_api_token,
                                            get_user_team_ids)
from rbintegrations.util.compat.logs import log_timed


logger = logging.getLogger(__name__)


class IDoneThisIntegration(Integration):
    """Integrates Review Board with I Done This.

    This integration allows posting 'done' entries to I Done This teams on
    behalf of users when they publish, change, or close a review request, and
    when they publish reviews or replies.

    I Done This entries are plain text, with #tags (used for group names)
    and @mentions (unused; only I Done This admins can configure them).
    URLs inside entries shown on the I Done This website are also
    automatically linked with truncated URL text.
    """

    name = _('I Done This')
    description = _(
        'Posts on behalf of users to I Done This teams when review requests '
        'are created, updated, and reviewed.'
    )

    default_settings = {
        'team_id': '',
    }

    config_form_cls = IDoneThisIntegrationConfigForm

    def initialize(self):
        """Initialize the integration hooks."""
        AccountPagesHook(self, [IDoneThisIntegrationAccountPage])

        hooks = (
            (review_request_closed, self._on_review_request_closed),
            (review_request_published, self._on_review_request_published),
            (review_request_reopened, self._on_review_request_reopened),
            (review_published, self._on_review_published),
            (reply_published, self._on_reply_published),
        )

        for signal, handler in hooks:
            SignalHook(self, signal, handler)

    @cached_property
    def icon_static_urls(self):
        """The icons used for the integration."""
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        return {
            '1x': extension.get_static_url('images/idonethis/icon.png'),
            '2x': extension.get_static_url('images/idonethis/icon@2x.png'),
        }

    def post_entry(self, entry_type, user, review_request, signal_name,
                   url=None, num_issues=0):
        """Post a 'done' entry to I Done This teams.

        Posts the specified entry to any configured teams that the user
        belongs to.

        Args:
            entry_type (unicode):
                Entry type from :py:mod:`rbintegrations.idonethis.entries` to
                post.

            user (django.contrib.auth.models.User):
                The user who updated the review request. Entries are only
                posted if the user has specified an API token for I Done This,
                and is a member of the configured teams.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request the notification is bound to.

            signal_name (unicode):
                The name of the signal triggering this notification.

            url (unicode, optional):
                URL to show in the entry instead of the review request URL.

            num_issues (int, optional):
                Number of issues opened in a review.
        """
        if not (review_request.public and user and user.is_active):
            return

        api_token = get_user_api_token(user)

        if not api_token:
            return

        user_team_ids = None

        for config in self.get_configs(review_request.local_site):
            if not config.match_conditions(form_cls=self.config_form_cls,
                                           review_request=review_request):
                continue

            # Lazy load team IDs after the first matching configuration.
            if user_team_ids is None:
                user_team_ids = get_user_team_ids(user)

            if not user_team_ids:
                # We finished posting to all of the user's teams, the request
                # to get the teams failed, or the user is not in any team.
                return

            team_id = config.get('team_id')

            if not team_id or team_id not in user_team_ids:
                continue

            # Avoid posting duplicate entries to the same team from multiple
            # matching configurations.
            user_team_ids.remove(team_id)

            template_string = entries.default_template_strings[entry_type]

            entry_body = entries.format_template_string(
                template_string=template_string,
                num_issues=num_issues,
                review_request=review_request,
                url=url)

            json_payload = json.dumps({
                'body': entry_body,
                'team_id': team_id,
                'status': 'done',
                # Optional 'occurred_on' is automatically set by I Done This.
            })

            request = create_idonethis_request(request_path='entries',
                                               api_token=api_token,
                                               json_payload=json_payload)

            with log_timed(f'Posting entry "{entry_type}" for signal '
                           f'"{signal_name}", review_request ID '
                           f'{review_request.pk}, user "{user.username}" to '
                           f'team "{team_id}", request '
                           f'"{request.get_method()} '
                           f'{request.get_full_url()}"',
                           logger=logger) as log_timer:
                try:
                    urlopen(request)
                except (HTTPError, URLError) as e:
                    # TODO: Record failure in user settings and possibly
                    #       notify the user on the account page so that
                    #       problems can be noticed.
                    if isinstance(e, HTTPError):
                        error_info = '%s, error data: %s' % (e, e.read())
                    else:
                        error_info = e.reason

                    logger.error('[%s] Failed to post entry for user "%s" '
                                 'to team "%s", request "%s %s": %s',
                                 log_timer.trace_id,
                                 user.username,
                                 team_id,
                                 request.get_method(),
                                 request.get_full_url(),
                                 error_info)
                except Exception as e:
                    logger.exception('[%s] Unexpected error posting entry '
                                     'for user "%s" to team "%s", request '
                                     '"%s %s": %s',
                                     log_timer.trace_id,
                                     user.username,
                                     team_id,
                                     request.get_method(),
                                     request.get_full_url(),
                                     e)

    def _on_review_request_closed(self, user, review_request, close_type,
                                  **kwargs):
        """Handler for when review requests are closed.

        Posts an entry to any configured I Done This teams when a review
        request is closed.

        Args:
            user (django.contrib.auth.models.User):
                The user who closed the review request.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request that was closed.

            close_type (unicode):
                The close type. Must be either ReviewRequest.DISCARDED or
                ReviewRequest.SUBMITTED.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.
        """
        if close_type == ReviewRequest.DISCARDED:
            entry_type = entries.REVIEW_REQUEST_DISCARDED
        else:
            entry_type = entries.REVIEW_REQUEST_COMPLETED

        self.post_entry(entry_type=entry_type,
                        user=user,
                        review_request=review_request,
                        signal_name='review_request_closed')

    def _on_review_request_published(self, user, review_request, changedesc,
                                     **kwargs):
        """Handler for when review requests are published.

        Posts an entry to any configured I Done This teams when a review
        request is published.

        Args:
            user (django.contrib.auth.models.User):
                The user who published the review request.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request that was published.

            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description for the update, if any.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.
        """
        if not changedesc:
            entry_type = entries.REVIEW_REQUEST_PUBLISHED
        elif ('status' in changedesc.fields_changed and
              changedesc.fields_changed['status']['new'][0] ==
                ReviewRequest.PENDING_REVIEW):
            entry_type = entries.REVIEW_REQUEST_REOPENED
        else:
            entry_type = entries.REVIEW_REQUEST_UPDATED

        self.post_entry(entry_type=entry_type,
                        user=user,
                        review_request=review_request,
                        signal_name='review_request_published')

    def _on_review_request_reopened(self, user, review_request, **kwargs):
        """Handler for when review requests are reopened.

        Posts an entry to any configured I Done This teams when a review
        request is reopened.

        Args:
            user (django.contrib.auth.models.User):
                The user who reopened the review request.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request that was reopened.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.
        """
        self.post_entry(entry_type=entries.REVIEW_REQUEST_REOPENED,
                        user=user,
                        review_request=review_request,
                        signal_name='review_request_reopened')

    def _on_review_published(self, user, review, to_owner_only, **kwargs):
        """Handler for when a review is published.

        Posts an entry to any configured I Done This teams when a review is
        published.

        Args:
            user (django.contrib.auth.models.User):
                The user who published the review.

            review (reviewboard.reviews.models.review.Review):
                The review that was published.

            to_owner_only (boolean):
                Whether the review should be sent only to the review request
                owner.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.
        """
        if to_owner_only:
            return

        num_issues = 0

        for comment in review.get_all_comments():
            if (comment.issue_opened and
                comment.issue_status == BaseComment.OPEN):
                num_issues += 1

        if review.ship_it:
            if num_issues == 0:
                entry_type = entries.REVIEW_PUBLISHED_SHIPIT
            elif num_issues == 1:
                entry_type = entries.REVIEW_PUBLISHED_SHIPIT_ISSUE
            else:
                entry_type = entries.REVIEW_PUBLISHED_SHIPIT_ISSUES
        else:
            if num_issues == 0:
                entry_type = entries.REVIEW_PUBLISHED
            elif num_issues == 1:
                entry_type = entries.REVIEW_PUBLISHED_ISSUE
            else:
                entry_type = entries.REVIEW_PUBLISHED_ISSUES

        self.post_entry(entry_type=entry_type,
                        user=user,
                        review_request=review.review_request,
                        signal_name='review_published',
                        url=build_server_url(review.get_absolute_url()),
                        num_issues=num_issues)

    def _on_reply_published(self, user, reply, **kwargs):
        """Handler for when a reply to a review is published.

        Posts an entry to any configured I Done This teams when a reply to a
        review is published.

        Args:
            user (django.contrib.auth.models.User):
                The user who published the reply.

            reply (reviewboard.reviews.models.review.Review):
                The reply that was published.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.
        """
        self.post_entry(entry_type=entries.REPLY_PUBLISHED,
                        user=user,
                        review_request=reply.review_request,
                        signal_name='reply_published',
                        url=build_server_url(reply.get_absolute_url()))
